from itertools import product
import json
import pytest
import re
from operator import attrgetter
from datetime import datetime, timezone, timedelta
from ...constants import *  # NOQA
from ...archiver.prune_cmd import (
    PRUNING_RULES,
    prune,
    PRUNE_DAILY,
    PRUNE_HOURLY,
    PRUNE_MINUTELY,
    PRUNE_MONTHLY,
    PRUNE_SECONDLY,
    PRUNE_WEEKLY,
    PRUNE_YEARLY,
)
from ...helpers import CommandError
from ...manifest import ArchiveInfo
from . import cmd, RK_ENCRYPTION, generate_archiver_tests

pytest_generate_tests = lambda metafunc: generate_archiver_tests(metafunc, kinds="local,binary")  # NOQA


def _create_archive_dt(archiver, backup_files, name, dt, tzinfo=timezone.utc):
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=tzinfo)
    cmd(archiver, "create", "--timestamp", dt.isoformat(timespec="microseconds"), name, backup_files)


def _create_archive_ts(archiver, backup_files, name, y, m, d, H=0, M=0, S=0, us=0, tzinfo=timezone.utc):
    _create_archive_dt(archiver, backup_files, name, datetime(y, m, d, H, M, S, us, tzinfo=tzinfo))


def test_prune_repository(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test1", backup_files)
    cmd(archiver, "create", "test2", backup_files)
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep-daily=1")
    assert re.search(r"Would prune:\s+test1", output)
    # Must keep the latest archive:
    assert re.search(r"Keeping archive \(rule: daily #1\):\s+test2", output)
    output = cmd(archiver, "repo-list")
    assert "test1" in output
    assert "test2" in output
    output = cmd(archiver, "prune", "--list", "--keep-daily=1")
    assert re.search(r"Pruning archive \(1/1\):\s+test1", output)
    output = cmd(archiver, "repo-list")
    assert "test1" not in output
    # The latest archive must still be there:
    assert "test2" in output


# This test must match docs/misc/prune-example.txt
def test_prune_repository_example(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    # Archives that will be kept, per the example
    # Oldest archive
    _create_archive_ts(archiver, backup_files, "test01", 2015, 1, 1)
    # 6 monthly archives
    _create_archive_ts(archiver, backup_files, "test02", 2015, 6, 30)
    _create_archive_ts(archiver, backup_files, "test03", 2015, 7, 31)
    _create_archive_ts(archiver, backup_files, "test04", 2015, 8, 31)
    _create_archive_ts(archiver, backup_files, "test05", 2015, 9, 30)
    _create_archive_ts(archiver, backup_files, "test06", 2015, 10, 31)
    _create_archive_ts(archiver, backup_files, "test07", 2015, 11, 30)
    # 14 daily archives
    _create_archive_ts(archiver, backup_files, "test08", 2015, 12, 17)
    _create_archive_ts(archiver, backup_files, "test09", 2015, 12, 18)
    _create_archive_ts(archiver, backup_files, "test10", 2015, 12, 20)
    _create_archive_ts(archiver, backup_files, "test11", 2015, 12, 21)
    _create_archive_ts(archiver, backup_files, "test12", 2015, 12, 22)
    _create_archive_ts(archiver, backup_files, "test13", 2015, 12, 23)
    _create_archive_ts(archiver, backup_files, "test14", 2015, 12, 24)
    _create_archive_ts(archiver, backup_files, "test15", 2015, 12, 25)
    _create_archive_ts(archiver, backup_files, "test16", 2015, 12, 26)
    _create_archive_ts(archiver, backup_files, "test17", 2015, 12, 27)
    _create_archive_ts(archiver, backup_files, "test18", 2015, 12, 28)
    _create_archive_ts(archiver, backup_files, "test19", 2015, 12, 29)
    _create_archive_ts(archiver, backup_files, "test20", 2015, 12, 30)
    _create_archive_ts(archiver, backup_files, "test21", 2015, 12, 31)
    # Additional archives that would be pruned
    # The second backup of the year
    _create_archive_ts(archiver, backup_files, "test22", 2015, 1, 2)
    # The next older monthly backup
    _create_archive_ts(archiver, backup_files, "test23", 2015, 5, 31)
    # The next older daily backup
    _create_archive_ts(archiver, backup_files, "test24", 2015, 12, 16)
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep-daily=14", "--keep-monthly=6", "--keep-yearly=1")
    # Prune second backup of the year
    assert re.search(r"Would prune:\s+test22", output)
    # Prune next older monthly and daily backups
    assert re.search(r"Would prune:\s+test23", output)
    assert re.search(r"Would prune:\s+test24", output)
    # Must keep the other 21 backups
    # Yearly is kept as oldest archive
    assert re.search(r"Keeping archive \(rule: yearly\[oldest\] #1\):\s+test01", output)
    for i in range(1, 7):
        assert re.search(r"Keeping archive \(rule: monthly #" + str(i) + r"\):\s+test" + ("%02d" % (8 - i)), output)
    for i in range(1, 15):
        assert re.search(r"Keeping archive \(rule: daily #" + str(i) + r"\):\s+test" + ("%02d" % (22 - i)), output)
    output = cmd(archiver, "repo-list")
    # Nothing pruned after dry run
    for i in range(1, 25):
        assert "test%02d" % i in output
    cmd(archiver, "prune", "--keep-daily=14", "--keep-monthly=6", "--keep-yearly=1")
    output = cmd(archiver, "repo-list")
    # All matching backups plus oldest kept
    for i in range(1, 22):
        assert "test%02d" % i in output
    # Other backups have been pruned
    for i in range(22, 25):
        assert "test%02d" % i not in output


# This test must match docs/misc/prune-example-interval.txt
def test_prune_repository_example_interval(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)

    # All timestamps are at exactly 16:00 UTC (matching the example).
    # Backups on most days from 2025-11-15 to 2026-06-04, with skips on
    # 2026-03-31, 2026-05-24, and 2026-06-03.
    archive_dates = [
        (2025, 11, 15),
        (2025, 12, 31),
        (2026, 1, 31),
        (2026, 2, 28),
        (2026, 3, 30),
        (2026, 4, 30),
        (2026, 5, 3),
        (2026, 5, 10),
        (2026, 5, 17),
        (2026, 5, 23),
        (2026, 5, 27),
        (2026, 5, 28),
        (2026, 5, 29),
        (2026, 5, 30),
        (2026, 5, 31),
        (2026, 6, 1),
        (2026, 6, 2),
        (2026, 6, 4),
    ]

    names = [f"backup_{y:04d}-{m:02d}-{d:02d}" for y, m, d in archive_dates]
    for (y, m, d), name in zip(archive_dates, names):
        _create_archive_ts(archiver, backup_files, name, y, m, d, H=16)

    output = cmd(
        archiver,
        "prune",
        "--list",
        "--dry-run",
        "--from=2026-06-04T16:00:00+00:00",
        "--keep-daily=1w",
        "--keep-weekly=4",
        "--keep-monthly=5m",
        "--keep-yearly=2",
    )

    # 2026-06-04 is kept unconditionally by the --from prefilter.
    assert re.search(r"Keeping archive \(rule: skip #1\):\s+backup_2026-06-04", output)

    daily_kept = [
        "backup_2026-06-02",
        "backup_2026-06-01",
        "backup_2026-05-31",
        "backup_2026-05-30",
        "backup_2026-05-29",
        "backup_2026-05-28",
    ]
    for i, name in enumerate(daily_kept, 1):
        assert re.search(rf"Keeping archive \(rule: daily #{i}\):\s+{name}", output)

    # Weekly W22 slot is consumed by 05-31 (already kept by daily),
    # so weekly reaches back to W18 to fill all 4 slots.
    weekly_kept = [
        "backup_2026-05-23",  # W21 — no Sunday candidate (05-24 skipped)
        "backup_2026-05-17",  # W20
        "backup_2026-05-10",  # W19
        "backup_2026-05-03",  # W18
    ]
    for i, name in enumerate(weekly_kept, 1):
        assert re.search(rf"Keeping archive \(rule: weekly #{i}\):\s+{name}", output)

    monthly_kept = [
        "backup_2026-04-30",
        "backup_2026-03-30",
        "backup_2026-02-28",
        "backup_2026-01-31",
        "backup_2025-12-31",
    ]
    for i, name in enumerate(monthly_kept, 1):
        assert re.search(rf"Keeping archive \(rule: monthly #{i}\):\s+{name}", output)

    # No true yearly candidates remain; only the oldest archive fills the slot.
    assert re.search(r"Keeping archive \(rule: yearly\[oldest\] #1\):\s+backup_2025-11-15", output)

    # 05-27 was kept by daily yesterday (window started 05-26) but falls out
    # today (window shifted to 05-28). W22 slot is consumed by daily-held 05-31,
    # so weekly doesn't save it either.
    assert re.search(r"Would prune:\s+backup_2026-05-27", output)


def test_prune_quarterly(archivers, request, backup_files):
    # Example worked through by hand when developing the quarterly
    # strategy, based on existing backups where the quarterly strategy
    # is desired. Weekly/monthly backups that do not affect results were
    # trimmed to speed up the test.
    #
    # The ISO week number is shown in a comment for each row in the list below.
    # The year is also shown when it does not match the year given in the
    # date tuple.
    archiver = request.getfixturevalue(archivers)
    test_dates = [
        (2020, 12, 6),
        (2021, 1, 3),  # 49, 2020-53
        (2021, 3, 28),
        (2021, 4, 25),  # 12, 16
        (2021, 6, 27),
        (2021, 7, 4),  # 25, 26
        (2021, 9, 26),
        (2021, 10, 3),  # 38, 39
        (2021, 12, 26),
        (2022, 1, 2),  # 51, 2021-52
    ]

    def mk_name(tup):
        (y, m, d) = tup
        suff = datetime(y, m, d).strftime("%Y-%m-%d")
        return f"test-{suff}"

    # The kept repos are based on working on an example by hand,
    # archives made on the following dates should be kept:
    EXPECTED_KEPT = {
        "13weekly": [(2020, 12, 6), (2021, 1, 3), (2021, 3, 28), (2021, 7, 4), (2021, 10, 3), (2022, 1, 2)],
        "3monthly": [(2020, 12, 6), (2021, 3, 28), (2021, 6, 27), (2021, 9, 26), (2021, 12, 26), (2022, 1, 2)],
    }

    for strat, to_keep in EXPECTED_KEPT.items():
        # Initialize our repo.
        cmd(archiver, "repo-create", RK_ENCRYPTION)
        for a, (y, m, d) in zip(map(mk_name, test_dates), test_dates):
            _create_archive_ts(archiver, backup_files, a, y, m, d)

        to_prune = list(set(test_dates) - set(to_keep))

        # Use 99 instead of -1 to test that oldest backup is kept.
        output = cmd(archiver, "prune", "--list", "--dry-run", f"--keep-{strat}=99")
        for a in map(mk_name, to_prune):
            assert re.search(rf"Would prune:\s+{a}", output)

        oldest = r"\[oldest\]" if strat in ("13weekly") else ""
        assert re.search(rf"Keeping archive \(rule: quarterly_{strat}{oldest} #\d+\):\s+test-2020-12-06", output)
        for a in map(mk_name, to_keep[1:]):
            assert re.search(rf"Keeping archive \(rule: quarterly_{strat} #\d+\):\s+{a}", output)

        output = cmd(archiver, "repo-list")
        # Nothing pruned after dry run
        for a in map(mk_name, test_dates):
            assert a in output

        cmd(archiver, "prune", f"--keep-{strat}=99")
        output = cmd(archiver, "repo-list")
        # All matching backups plus oldest kept
        for a in map(mk_name, to_keep):
            assert a in output
        # Other backups have been pruned
        for a in map(mk_name, to_prune):
            assert a not in output

        # Delete repo and begin anew
        cmd(archiver, "repo-delete")


# With an initial and daily backup, prune daily until oldest is replaced by a monthly backup
def test_prune_retain_and_expire_oldest(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    # Initial backup
    _create_archive_ts(archiver, backup_files, "original_archive", 2020, 9, 1, 11, 15)
    # Archive and prune daily for 30 days
    for i in range(1, 31):
        _create_archive_ts(archiver, backup_files, "september%02d" % i, 2020, 9, i, 12)
        cmd(archiver, "prune", "--keep-daily=7", "--keep-monthly=1")
    # Archive and prune 6 days into the next month
    for i in range(1, 7):
        _create_archive_ts(archiver, backup_files, "october%02d" % i, 2020, 10, i, 12)
        cmd(archiver, "prune", "--keep-daily=7", "--keep-monthly=1")
    # Oldest backup is still retained
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep-daily=7", "--keep-monthly=1")
    assert re.search(r"Keeping archive \(rule: monthly\[oldest\] #1" + r"\):\s+original_archive", output)
    # Archive one more day and prune.
    _create_archive_ts(archiver, backup_files, "october07", 2020, 10, 7, 12)
    cmd(archiver, "prune", "--keep-daily=7", "--keep-monthly=1")
    # Last day of previous month is retained as monthly, and oldest is expired.
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep-daily=7", "--keep-monthly=1")
    assert re.search(r"Keeping archive \(rule: monthly #1\):\s+september30", output)
    assert "original_archive" not in output


def test_prune_repository_prefix(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "foo-2015-08-12-10:00", backup_files)
    cmd(archiver, "create", "foo-2015-08-12-20:00", backup_files)
    cmd(archiver, "create", "bar-2015-08-12-10:00", backup_files)
    cmd(archiver, "create", "bar-2015-08-12-20:00", backup_files)
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep-daily=1", "--match-archives=sh:foo-*")
    assert re.search(r"Keeping archive \(rule: daily #1\):\s+foo-2015-08-12-20:00", output)
    assert re.search(r"Would prune:\s+foo-2015-08-12-10:00", output)
    output = cmd(archiver, "repo-list")
    assert "foo-2015-08-12-10:00" in output
    assert "foo-2015-08-12-20:00" in output
    assert "bar-2015-08-12-10:00" in output
    assert "bar-2015-08-12-20:00" in output
    cmd(archiver, "prune", "--keep-daily=1", "--match-archives=sh:foo-*")
    output = cmd(archiver, "repo-list")
    assert "foo-2015-08-12-10:00" not in output
    assert "foo-2015-08-12-20:00" in output
    assert "bar-2015-08-12-10:00" in output
    assert "bar-2015-08-12-20:00" in output


def test_prune_repository_glob(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "2015-08-12-10:00-foo", backup_files)
    cmd(archiver, "create", "2015-08-12-20:00-foo", backup_files)
    cmd(archiver, "create", "2015-08-12-10:00-bar", backup_files)
    cmd(archiver, "create", "2015-08-12-20:00-bar", backup_files)
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep-daily=1", "--match-archives=sh:2015-*-foo")
    assert re.search(r"Keeping archive \(rule: daily #1\):\s+2015-08-12-20:00-foo", output)
    assert re.search(r"Would prune:\s+2015-08-12-10:00-foo", output)
    output = cmd(archiver, "repo-list")
    assert "2015-08-12-10:00-foo" in output
    assert "2015-08-12-20:00-foo" in output
    assert "2015-08-12-10:00-bar" in output
    assert "2015-08-12-20:00-bar" in output
    cmd(archiver, "prune", "--keep-daily=1", "--match-archives=sh:2015-*-foo")
    output = cmd(archiver, "repo-list")
    assert "2015-08-12-10:00-foo" not in output
    assert "2015-08-12-20:00-foo" in output
    assert "2015-08-12-10:00-bar" in output
    assert "2015-08-12-20:00-bar" in output


def test_prune_ignore_protected(archivers, request):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "archive1", archiver.input_path)
    cmd(archiver, "tag", "--set=@PROT", "archive1")  # do not delete archive1!
    cmd(archiver, "create", "archive2", archiver.input_path)
    cmd(archiver, "create", "archive3", archiver.input_path)
    output = cmd(archiver, "prune", "--list", "--keep=1", "--match-archives=sh:archive*")
    assert "archive1" not in output  # @PROT archives are completely ignored.
    assert re.search(r"Keeping archive \(rule: keep #1\):\s+archive3", output)
    assert re.search(r"Pruning archive \(.*?\):\s+archive2", output)
    output = cmd(archiver, "repo-list")
    assert "archive1" in output  # @PROT protected archive1 from deletion
    assert "archive3" in output  # last one


mock_id = 0


def mock_archive(ts, id=None):
    """Create an ArchiveInfo with mocked/default values."""
    global mock_id
    if id is None:
        id = mock_id
        mock_id += 1
    return ArchiveInfo(name="", id=id, ts=ts.replace(tzinfo=timezone.utc), tags=(), host="", user="")


@pytest.mark.parametrize(
    "rule,num_to_keep,expected_indices",
    [
        (PRUNE_YEARLY, 3, (12, 1, 0)),
        (PRUNE_MONTHLY, 3, (12, 7, 3)),
        (PRUNE_WEEKLY, 2, (12, 7)),
        (PRUNE_DAILY, 3, (12, 7, 6)),
        (PRUNE_HOURLY, 3, (12, 9, 7)),
        (PRUNE_MINUTELY, 3, (12, 9, 8)),
        (PRUNE_SECONDLY, 4, (12, 11, 10, 9)),
        (PRUNE_DAILY, 0, []),
        (PRUNE_DAILY, -1, (12, 7, 6, 5, 4, 3, 2, 1, 0)),
    ],
)
def test_prune(rule, num_to_keep, expected_indices):
    archives = [
        # years apart
        mock_archive(datetime(2015, 1, 1, 10, 0, 0)),
        mock_archive(datetime(2016, 1, 1, 10, 0, 0)),
        mock_archive(datetime(2017, 1, 1, 10, 0, 0)),
        # months apart
        mock_archive(datetime(2017, 2, 1, 10, 0, 0)),
        mock_archive(datetime(2017, 3, 1, 10, 0, 0)),
        # days apart
        mock_archive(datetime(2017, 3, 2, 10, 0, 0)),
        mock_archive(datetime(2017, 3, 3, 10, 0, 0)),
        mock_archive(datetime(2017, 3, 4, 10, 0, 0)),
        # minutes apart
        mock_archive(datetime(2017, 10, 1, 9, 45, 0)),
        mock_archive(datetime(2017, 10, 1, 9, 55, 0)),
        # seconds apart
        mock_archive(datetime(2017, 10, 1, 10, 0, 1)),
        mock_archive(datetime(2017, 10, 1, 10, 0, 3)),
        mock_archive(datetime(2017, 10, 1, 10, 0, 5)),
    ]
    keep = prune(sorted(archives, key=attrgetter("ts"), reverse=True), rule, num_to_keep, None, False)

    assert set(keep) == {archives[i] for i in expected_indices}
    assert all(result.rule == rule for _, result in keep.items())


def test_prune_keep_oldest():
    archives = [
        # oldest backup, but not last in its year
        mock_archive(datetime(2018, 1, 1, 10, 0, 0)),
        # an interim backup
        mock_archive(datetime(2018, 12, 30, 10, 0, 0)),
        # year-end backups
        mock_archive(datetime(2018, 12, 31, 10, 0, 0)),
        mock_archive(datetime(2019, 12, 31, 10, 0, 0)),
    ]
    sorted_archives = sorted(archives, key=attrgetter("ts"), reverse=True)

    # Keep oldest when retention target can't otherwise be met
    keep = prune(sorted_archives, PRUNE_YEARLY, 3, None, True)

    assert keep[archives[0]].rule.key == "yearly" and keep[archives[0]].oldest is True
    assert keep[archives[2]].rule.key == "yearly" and keep[archives[2]].oldest is False
    assert keep[archives[3]].rule.key == "yearly" and keep[archives[3]].oldest is False
    assert len(keep) == 3

    # Otherwise, prune it
    keep = prune(sorted_archives, PRUNE_YEARLY, 2, None, True)

    assert keep[archives[2]].rule.key == "yearly" and keep[archives[2]].oldest is False
    assert keep[archives[3]].rule.key == "yearly" and keep[archives[3]].oldest is False
    assert len(keep) == 2


def test_prune_no_archives():
    archives = []

    keep = prune(archives, PRUNE_YEARLY, 3, None, False)

    assert keep == {}


def test_prune_list_with_metadata_format(archivers, request, backup_files):
    # Regression test for: prune --list with a format string that requires loading
    # archive metadata (e.g. {hostname}) must not fail when archives are deleted.
    # The bug was that format_item() was called after archive.delete(), causing
    # Archive.DoesNotExist when the formatter tried to lazy-load the archive.
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test1", backup_files)
    cmd(archiver, "create", "test2", backup_files)
    # {hostname} is a "call key" that triggers lazy loading of the archive from the repo.
    # With the buggy code this would raise Archive.DoesNotExist for the pruned archive.
    output = cmd(archiver, "prune", "--list", "--keep-daily=1", "--format={name} {hostname}{NL}")
    assert "test1" in output
    assert "test2" in output


def test_prune_json(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test1", backup_files)
    cmd(archiver, "create", "test2", backup_files)
    prune_result = json.loads(cmd(archiver, "prune", "--json", "--dry-run", "--keep-daily=1"))
    assert "repository" in prune_result
    assert "encryption" in prune_result
    assert len(prune_result["repository"]["id"]) == 64
    archives = prune_result["archives"]
    assert len(archives) == 2
    kept = [a for a in archives if a["kept"]]
    pruned = [a for a in archives if not a["kept"]]
    assert len(kept) == 1
    assert len(pruned) == 1
    assert kept[0]["name"] == "test2"
    assert kept[0]["keep_rule"] == "daily"
    assert kept[0]["kept_archive_number"] == 1
    assert not kept[0]["kept_oldest"]
    assert "deleted_archive_number" not in kept[0]
    assert pruned[0]["name"] == "test1"
    assert pruned[0]["deleted_archive_number"] == 1
    assert "keep_rule" not in pruned[0]
    assert "kept_archive_number" not in pruned[0]
    for archive in archives:
        assert "name" in archive
        assert "id" in archive
        assert "time" in archive
        assert "kept" in archive


def test_prune_json_list_pruned(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test1", backup_files)
    cmd(archiver, "create", "test2", backup_files)
    prune_result = json.loads(cmd(archiver, "prune", "--json", "--dry-run", "--list-pruned", "--keep-daily=1"))
    archives = prune_result["archives"]
    assert len(archives) == 1
    assert archives[0]["name"] == "test1"
    assert archives[0]["kept"] is False
    assert archives[0]["deleted_archive_number"] == 1


def test_prune_keep_same_second(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    cmd(archiver, "create", "test1", backup_files)
    cmd(archiver, "create", "test2", backup_files)
    output = cmd(archiver, "prune", "--list", "--dry-run", "--keep=2")
    # Both archives are kept even though they have the same timestamp to the second. Would previously have failed with
    # old behavior of --keep-last. Archives sorted on seconds, order is undefined.
    assert re.search(r"Keeping archive \(rule: keep #\d\):\s+test1", output)
    assert re.search(r"Keeping archive \(rule: keep #\d\):\s+test2", output)


@pytest.mark.parametrize("keep_arg", ["--keep=2", "--keep=1S"])
def test_prune_keep_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    # Two archives at the --from since boundary prove the inclusive timestamp check.
    _create_archive_dt(archiver, backup_files, "test-1", dt)
    _create_archive_dt(
        archiver, backup_files, "test-2", dt - timedelta(microseconds=999999)
    )  # Would be pruned if `secondly`-rule was active.
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(seconds=1))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(seconds=1, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: skip #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: keep #1\):\s+test-2", output)
    assert re.search(r"Keeping archive \(rule: keep #2\):\s+test-3", output)
    assert re.search(r"Would prune:\s+test-4", output)


@pytest.mark.parametrize("keep_arg", ["--keep-secondly=2", "--keep-secondly=2S"])
def test_prune_keep_secondly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(seconds=1))
    _create_archive_dt(archiver, backup_files, "test-2", dt - timedelta(seconds=1, microseconds=999999))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(seconds=2))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(seconds=2, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: secondly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: secondly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Would prune:\s+test-4", output)


@pytest.mark.parametrize("keep_arg", ["--keep-minutely=3", "--keep-minutely=3M"])
def test_prune_keep_minutely_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 23, 59, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(minutes=1))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(minutes=2)).replace(second=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(minutes=2))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(minutes=3))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(minutes=3, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: minutely #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: minutely #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: minutely #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-hourly=3", "--keep-hourly=3H"])
def test_prune_keep_hourly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 23, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(hours=1))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(hours=2)).replace(microsecond=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(hours=2))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(hours=3))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(hours=3, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: hourly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: hourly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: hourly #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-daily=3", "--keep-daily=3d"])
def test_prune_keep_daily_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 0, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=1))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(days=2)).replace(second=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=2))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=3))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(days=3, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: daily #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: daily #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: daily #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-weekly=3", "--keep-weekly=3w"])
def test_prune_keep_weekly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 0, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=7))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(days=14)).replace(microsecond=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=14))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=21))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(days=21, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: weekly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: weekly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: weekly #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-monthly=3", "--keep-monthly=3m"])
def test_prune_keep_monthly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 0, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=31))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(days=62)).replace(microsecond=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=62))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=93))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(days=93, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: monthly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: monthly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: monthly #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


# 2023-12-31 is Sunday, week 52. Makes these week calculations a little easier.
@pytest.mark.parametrize("keep_arg", ["--keep-13weekly=3", "--keep-13weekly=39w"])
def test_prune_keep_13weekly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 0, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=91))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(days=182)).replace(microsecond=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=182))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=273))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(days=273, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: quarterly_13weekly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: quarterly_13weekly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: quarterly_13weekly #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-3monthly=3", "--keep-3monthly=275d"])
def test_prune_keep_3monthly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 0, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=92))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(days=184)).replace(microsecond=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=184))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=275))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(days=275, microseconds=1))
    # 275d is the interval from dt to the oldest kept monthly archive
    output = cmd(archiver, "prune", "--list", "--short", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: quarterly_3monthly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: quarterly_3monthly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: quarterly_3monthly #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-yearly=3", "--keep-yearly=3y"])
def test_prune_keep_yearly_int_or_interval(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 0, 0, 0, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=365))
    _create_archive_dt(archiver, backup_files, "test-2", (dt - timedelta(days=730)).replace(microsecond=1))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=730))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=1095))
    _create_archive_dt(archiver, backup_files, "test-5", dt - timedelta(days=1095, microseconds=1))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: yearly #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: yearly #2\):\s+test-2", output)
    assert re.search(r"Would prune:\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: yearly #3\):\s+test-4", output)
    assert re.search(r"Would prune:\s+test-5", output)


@pytest.mark.parametrize("keep_arg", ["--keep-daily=-1", "--keep-daily=all"])
def test_prune_keep_daily_all(archivers, request, backup_files, keep_arg):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(days=1))
    _create_archive_dt(archiver, backup_files, "test-2", dt - timedelta(days=2))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=3))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: daily #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: daily #2\):\s+test-2", output)
    assert re.search(r"Keeping archive \(rule: daily #3\):\s+test-3", output)


@pytest.mark.parametrize("keep_arg", ["--keep=-1", "--keep=all"])
def test_prune_keep_flat_all(archivers, request, backup_files, keep_arg):
    """--keep=all / --keep=-1 keeps every archive."""
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2023, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    _create_archive_dt(archiver, backup_files, "test-1", dt - timedelta(microseconds=1))
    _create_archive_dt(archiver, backup_files, "test-2", dt - timedelta(microseconds=2))
    _create_archive_dt(archiver, backup_files, "test-3", dt - timedelta(days=3))
    _create_archive_dt(archiver, backup_files, "test-4", dt - timedelta(days=3333))
    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), keep_arg)
    assert re.search(r"Keeping archive \(rule: keep #1\):\s+test-1", output)
    assert re.search(r"Keeping archive \(rule: keep #2\):\s+test-2", output)
    assert re.search(r"Keeping archive \(rule: keep #3\):\s+test-3", output)
    assert re.search(r"Keeping archive \(rule: keep #4\):\s+test-4", output)


def _cmd_prune_error(archiver, *args):
    """Run prune expecting a CommandError. Returns error string for assertions."""
    if archiver.FORK_DEFAULT:
        return cmd(archiver, "prune", *args, exit_code=CommandError().exit_code)
    with pytest.raises(CommandError) as error:
        cmd(archiver, "prune", *args)
    return str(error.value)


def test_prune_no_args(archivers, request):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    output = _cmd_prune_error(archiver)

    assert re.search(r"At least one of the .* settings must be specified.", output)
    assert re.search(r"keep(?!-)", output)
    flags = ["secondly", "minutely", "hourly", "daily", "weekly", "monthly", "yearly", "13weekly", "3monthly"]
    for flag in flags:
        assert f"keep-{flag}" in output


@pytest.mark.parametrize("keep_arg,value", list(product([rule.key for rule in PRUNING_RULES], ["0", "0S"])))
def test_prune_all_zero_args_one(archivers, request, keep_arg, value):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)

    arg_with_prefix = "--keep" if keep_arg == "keep" else f"--keep-{keep_arg.replace('quarterly_', '')}"
    output = _cmd_prune_error(archiver, f"{arg_with_prefix}={value}")
    assert re.search(r"None of the .* settings have a positive value. At least one must be non-zero.", output)


def test_prune_all_zero_multiple_multiple(archivers, request):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)

    output = _cmd_prune_error(archiver, "--keep-secondly=0S", "--keep-daily=0")
    assert re.search(r"None of the .* settings have a positive value. At least one must be non-zero.", output)


@pytest.mark.parametrize(
    "lo_val,hi_val",
    [("14d", "7d"), ("-1", "7d"), ("-1", "1"), ("-1", "-1"), ("all", "7d"), ("all", "1"), ("all", "-1")],
)
def test_prune_warns_on_redundant_interval_flags(archivers, request, lo_val, hi_val):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    output = _cmd_prune_error(archiver, "--dry-run", f"--keep-hourly={lo_val}", f"--keep-daily={hi_val}")
    assert "hourly=" in output
    assert "daily=" in output
    assert "effectively useless" in output


@pytest.mark.parametrize("lo_val,hi_val", [("7d", "14d"), ("7d", "-1"), ("1", "-1"), ("7d", "all"), ("1", "all")])
def test_prune_does_not_warn_on_normal_interval_flags(archivers, request, lo_val, hi_val):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    output = cmd(archiver, "prune", "--dry-run", f"--keep-hourly={lo_val}", f"--keep-daily={hi_val}")
    assert "effectively useless" not in output


def test_prune_int_rolling_schedule_oldest_retention():
    daily_n = 6
    monthly_n = 3
    start_date = datetime(2024, 1, 1, tzinfo=timezone.utc)

    previous_archives = []
    archives = []

    for day_offset in range(97):
        backup_ts = start_date + timedelta(days=day_offset)
        previous_archives = archives
        archives = [mock_archive(backup_ts, day_offset), *archives]

        keep = {}
        keep |= prune(archives, PRUNE_DAILY, daily_n, None, False, keep)
        keep |= prune(archives, PRUNE_MONTHLY, monthly_n, None, True, keep)

        archives = sorted(keep.keys(), key=lambda a: a.ts, reverse=True)

    # It is now 2024-04-06. The last run should have just pruned the jan-01
    # archive since the monthly retention count is now satisfied at jan-31. It
    # was kept until now to satisfy the oldest-rule.

    assert previous_archives[-1].ts.strftime("%m-%d") == "01-01"
    assert archives[-1].ts.strftime("%m-%d") == "01-31"


def test_prune_from_prefiltered_archives_ignored_in_pruning(archivers, request, backup_files):
    archiver = request.getfixturevalue(archivers)
    cmd(archiver, "repo-create", RK_ENCRYPTION)
    dt = datetime(2024, 6, 6, 12, 0, 0, tzinfo=timezone.utc)

    _create_archive_dt(archiver, backup_files, "test-a", dt + timedelta(hours=1))
    _create_archive_dt(archiver, backup_files, "test-b", dt - timedelta(hours=1))
    _create_archive_dt(archiver, backup_files, "test-c", dt - timedelta(days=1))

    output = cmd(archiver, "prune", "--list", "--dry-run", "--from", dt.isoformat(), "--keep-daily=1")

    # 'test-b' is kept, meaning 'test-a' was entirely skipped for pruning consideration.
    # They would otherwise have occupied the same period.
    assert re.search(r"Keeping archive \(rule: skip #1\):\s+test-a", output)
    assert re.search(r"Keeping archive \(rule: daily #1\):\s+test-b", output)

    assert re.search(r"Would prune:\s+test-c", output)


def test_prune_interval_rolling_schedule_oldest_retention():
    daily_interval = timedelta(days=6)
    monthly_interval = timedelta(days=31 * 3)  # Matching --keep-monthly=3m after argument parsing
    start_date = datetime(2024, 1, 1, tzinfo=timezone.utc)

    previous_archives = []
    archives = []

    for day_offset in range(95):
        backup_ts = start_date + timedelta(days=day_offset)
        previous_archives = archives
        archives = [mock_archive(backup_ts, day_offset), *archives]

        keep = {}
        keep |= prune(archives, PRUNE_DAILY, daily_interval, backup_ts, False, keep)
        keep |= prune(archives, PRUNE_MONTHLY, monthly_interval, backup_ts, True, keep)

        print(
            f"For backup+prune at {backup_ts.strftime('%m-%d')} ({day_offset})"
            f" the following {len(archives)} archives are kept:"
        )
        for a, result in keep.items():
            print(f"    {a.id}: {a.ts.strftime('%Y-%m-%d')} {result}")

        archives = sorted(keep.keys(), key=lambda a: a.ts, reverse=True)

    # It is now 2024-04-04. On this final run jan-01 falls outside the
    # 93d monthly interval and is pruned. It was kept until now by the
    # oldest-retention rule — on 2024-04-03 it was exactly at the
    # inclusive boundary and retained as oldest.

    assert previous_archives[-1].ts.strftime("%m-%d") == "01-01"
    assert archives[-1].ts.strftime("%m-%d") == "01-31"



"""
This package contains helper and utility functionality that did not fit elsewhere.

Code used to be in borg/helpers.py but was split into modules in this
package, which are imported here for compatibility.
"""

import os
import logging
from collections import namedtuple

from ..constants import *  # NOQA

from .datastruct import StableDict, Buffer, EfficientCollectionQueue
from .errors import Error, ErrorWithTraceback, IntegrityError, DecompressionError, CancelledByUser, CommandError
from .errors import RTError, PathNotAllowed, modern_ec
from .errors import BorgWarning, FileChangedWarning, BackupWarning, IncludePatternNeverMatchedWarning
from .errors import BackupError, BackupOSError, BackupRaceConditionError, BackupItemExcluded
from .errors import BackupPermissionError, BackupIOError, BackupFileNotFoundError
from .errors import BackupSymlinkParentError, BackupPathTraversalError, BackupHardlinkSourceError
from .fs import ensure_dir, join_base_dir
from .fs import get_security_dir, get_keys_dir, get_base_dir, get_cache_dir, get_config_dir, get_runtime_dir
from .fs import dir_is_tagged, dir_is_cachedir, remove_dotdot_prefixes, make_path_safe, scandir_inorder
from .fs import secure_erase, safe_unlink, dash_open, os_open, os_stat, get_strip_prefix, umount, slashify
from .fs import O_, flags_dir, flags_special_follow, flags_special, flags_base, flags_normal, flags_noatime
from .fs import HardLinkManager
from .misc import sysinfo, log_multi, consume
from .misc import ChunkIteratorFileWrapper, open_item, chunkit, iter_separated, ErrorIgnoringTextIOWrapper
from .parseformat import octal_int, bin_to_hex, hex_to_bin, safe_encode, safe_decode
from .parseformat import text_to_json, binary_to_json, remove_surrogates, join_cmd
from .parseformat import eval_escapes, decode_dict, interval, int_or_interval
from .parseformat import (
    PathSpec,
    FilesystemPathSpec,
    SortBySpec,
    CompressionSpec,
    ChunkerParams,
    FilesCacheMode,
    partial_format,
    DatetimeWrapper,
)
from .parseformat import format_file_size, parse_file_size, FileSize
from .parseformat import sizeof_fmt, sizeof_fmt_iec, sizeof_fmt_decimal, Location, text_validator
from .parseformat import format_line, replace_placeholders, PlaceholderError, relative_time_marker_validator
from .parseformat import format_archive, parse_stringified_list, clean_lines
from .parseformat import location_validator, archivename_validator, comment_validator, tag_validator
from .parseformat import BaseFormatter, ArchiveFormatter, ItemFormatter, DiffFormatter, file_status
from .parseformat import swidth_slice, ellipsis_truncate
from .parseformat import BorgJsonEncoder, basic_json_data, json_print, json_dump, prepare_dump_dict
from .parseformat import Highlander, MakePathSafeAction
from .process import daemonize, daemonizing, ThreadRunner
from .process import signal_handler, raising_signal_handler, sig_int, ignore_sigint, SigHup, SigTerm
from .process import popen_with_error_handling, is_terminal, prepare_subprocess_env, create_filter_process
from .progress import ProgressIndicatorPercent, ProgressIndicatorMessage
from .time import parse_timestamp, timestamp, safe_timestamp, safe_s, safe_ns, MAX_S, SUPPORT_32BIT_PLATFORMS
from .time import format_time, format_timedelta, OutputTimestamp, archive_ts_now
from .yes_no import yes, TRUISH, FALSISH, DEFAULTISH

from .msgpack import is_slow_msgpack, is_supported_msgpack, get_limited_unpacker
from . import msgpack

from ..logger import create_logger

logger = create_logger()


# generic mechanism to enable users to invoke workarounds by setting the
# BORG_WORKAROUNDS environment variable to a list of comma-separated strings.
# see the docs for a list of known workaround strings.
workarounds = tuple(os.environ.get("BORG_WORKAROUNDS", "").split(","))


# element data type for warnings_list:
warning_info = namedtuple("warning_info", "wc,msg,args,wt")

"""
The global warnings_list variable is used to collect warning_info elements while Borg is running.
"""
_warnings_list: list[warning_info] = []


def add_warning(msg, *args, **kwargs):
    global _warnings_list
    warning_code = kwargs.get("wc", EXIT_WARNING)
    assert isinstance(warning_code, int)
    warning_type = kwargs.get("wt", "percent")
    assert warning_type in ("percent", "curly")
    _warnings_list.append(warning_info(warning_code, msg, args, warning_type))


"""
The global exit_code variable allows modules other than archiver to increase the program's exit code if a
warning or error occurs during their operation.
"""
_exit_code = EXIT_SUCCESS


def classify_ec(ec):
    if not isinstance(ec, int):
        raise TypeError("ec must be of type int")
    if EXIT_SIGNAL_BASE <= ec <= 255:
        return "signal"
    elif ec == EXIT_ERROR or EXIT_ERROR_BASE <= ec < EXIT_WARNING_BASE:
        return "error"
    elif ec == EXIT_WARNING or EXIT_WARNING_BASE <= ec < EXIT_SIGNAL_BASE:
        return "warning"
    elif ec == EXIT_SUCCESS:
        return "success"
    else:
        raise ValueError(f"invalid error code: {ec}")


def max_ec(ec1, ec2):
    """Return the more severe error code of ec1 and ec2."""
    # note: usually, there can be only 1 error-class ec, the other ec is then either success or warning.
    ec1_class = classify_ec(ec1)
    ec2_class = classify_ec(ec2)
    if ec1_class == "signal":
        return ec1
    if ec2_class == "signal":
        return ec2
    if ec1_class == "error":
        return ec1
    if ec2_class == "error":
        return ec2
    if ec1_class == "warning":
        return ec1
    if ec2_class == "warning":
        return ec2
    assert ec1 == ec2 == EXIT_SUCCESS
    return EXIT_SUCCESS


def set_ec(ec):
    """
    Set the exit code of the program to ec if ec is more severe than the current exit code.
    """
    global _exit_code
    _exit_code = max_ec(_exit_code, ec)


def init_ec_warnings(ec=EXIT_SUCCESS, warnings=None):
    """
    (Re-)initialize the globals for the exit code and the warnings list.
    """
    global _exit_code, _warnings_list
    _exit_code = ec
    warnings = [] if warnings is None else warnings
    assert isinstance(warnings, list)
    _warnings_list = warnings


def get_ec(ec=None):
    """
    Compute the final return code of the Borg process.
    """
    if ec is not None:
        set_ec(ec)

    global _exit_code
    exit_code_class = classify_ec(_exit_code)
    if exit_code_class in ("signal", "error", "warning"):
        # there was a signal/error/warning, return its exit code
        return _exit_code
    assert exit_code_class == "success"
    global _warnings_list
    if not _warnings_list:
        # we do not have any warnings in warnings list, return success exit code
        return _exit_code
    # looks like we have some warning(s)
    rcs = sorted({w_info.wc for w_info in _warnings_list})
    logger.debug(f"rcs: {rcs!r}")
    if len(rcs) == 1:
        # easy: there was only one kind of warning, so we can be specific
        return rcs[0]
    # there were different kinds of warnings
    return EXIT_WARNING  # generic warning rc, user has to look into the logs


def get_reset_ec(ec=None):
    """Like get_ec, but reinitialize ec/warnings afterwards."""
    rc = get_ec(ec)
    init_ec_warnings()
    return rc


def do_show_rc(exit_code):
    """Log the program return code using the dedicated 'borg.output.show-rc' logger.

    Uses INFO/WARNING/ERROR levels depending on the classified exit code.

    This helper is robust: it swallows any exceptions to avoid interfering with
    program exit behavior. Callers do not need to guard it with try/except.
    """
    try:
        exit_msg = "terminating with %s status, rc %d"
        rc_logger = logging.getLogger("borg.output.show-rc")
        try:
            ec_class = classify_ec(exit_code)
        except ValueError:
            rc_logger.error(exit_msg % ("abnormal", exit_code or 666))
        else:
            if ec_class == "success":
                rc_logger.info(exit_msg % (ec_class, exit_code))
            elif ec_class == "warning":
                rc_logger.warning(exit_msg % (ec_class, exit_code))
            elif ec_class in ("error", "signal"):
                rc_logger.error(exit_msg % (ec_class, exit_code))
    except Exception:  # nosec B110
        # Never let logging issues interfere with exit behaviour
        pass



import time
from pathlib import Path

import pytest

from borgstore.store import Store

from ..storelocking import Lock, NotLocked, LockTimeout

ID1 = "foo", 1, 1
ID2 = "bar", 2, 2


@pytest.fixture()
def lockstore(tmp_path):
    store = Store(Path(tmp_path / "lockstore").as_uri(), config={"locks/": {"levels": [0]}})
    store.create()
    with store:
        yield store
    store.destroy()


class TestLock:
    def test_cm(self, lockstore):
        with Lock(lockstore, exclusive=True, id=ID1) as lock:
            assert lock.got_exclusive_lock()
        with Lock(lockstore, exclusive=False, id=ID1) as lock:
            assert not lock.got_exclusive_lock()

    def test_got_exclusive_lock(self, lockstore):
        lock = Lock(lockstore, exclusive=True, id=ID1)
        assert not lock.got_exclusive_lock()
        lock.acquire()
        assert lock.got_exclusive_lock()
        lock.release()
        assert not lock.got_exclusive_lock()

    def test_exclusive_lock(self, lockstore):
        # There must not be two exclusive locks.
        with Lock(lockstore, exclusive=True, id=ID1):
            with pytest.raises(LockTimeout):
                Lock(lockstore, exclusive=True, id=ID2).acquire()
        # Acquiring an exclusive lock will time out if the non-exclusive lock does not go away.
        with Lock(lockstore, exclusive=False, id=ID1):
            with pytest.raises(LockTimeout):
                Lock(lockstore, exclusive=True, id=ID2).acquire()

    def test_exclusive_lock_timeout_leaves_no_lock(self, lockstore):
        # When acquiring an exclusive lock times out because a non-exclusive lock does not go away,
        # the not-acquired exclusive lock must not stay behind in the store: it would block all
        # other clients (even on other hosts) until it expired as stale.
        with Lock(lockstore, exclusive=False, id=ID1) as shared_lock:
            with pytest.raises(LockTimeout):
                Lock(lockstore, exclusive=True, id=ID2).acquire()
            locks = shared_lock._get_locks()
            assert len(locks) == 1  # only the non-exclusive lock of ID1 is left
            assert not any(lock["exclusive"] for lock in locks.values())

    def test_double_nonexclusive_lock_succeeds(self, lockstore):
        with Lock(lockstore, exclusive=False, id=ID1):
            with Lock(lockstore, exclusive=False, id=ID2):
                pass

    def test_not_locked(self, lockstore):
        lock = Lock(lockstore, exclusive=True, id=ID1)
        with pytest.raises(NotLocked):
            lock.release()
        lock = Lock(lockstore, exclusive=False, id=ID1)
        with pytest.raises(NotLocked):
            lock.release()

    def test_break_lock(self, lockstore):
        lock = Lock(lockstore, exclusive=True, id=ID1).acquire()
        lock.break_lock()
        with Lock(lockstore, exclusive=True, id=ID2):
            pass
        with Lock(lockstore, exclusive=True, id=ID1):
            pass

    def test_lock_refresh_stale_removal(self, lockstore):
        # stale after 2s, refreshable after 1s
        lock = Lock(lockstore, exclusive=True, id=ID1, stale=2)
        lock.acquire()
        lock_keys_a00 = set(lock._get_locks())
        time.sleep(0.5)
        lock.refresh()  # Should not change locks; existing lock is too young.
        lock_keys_a05 = set(lock._get_locks())
        time.sleep(0.6)
        lock.refresh()  # This should refresh the lock.
        lock_keys_b00 = set(lock._get_locks())
        time.sleep(2.1)
        # now the lock is stale. we never consider the lock we hold ourselves stale,
        # but another client (== another Lock instance) does:
        other_lock = Lock(lockstore, exclusive=True, id=ID2, stale=2)
        lock_keys_b21 = set(other_lock._get_locks())  # now the lock should be stale & gone.
        assert lock_keys_a00 == lock_keys_a05  # was too young, no refresh done
        assert len(lock_keys_a00) == 1
        assert lock_keys_a00 != lock_keys_b00  # refresh done, new lock has different key
        assert len(lock_keys_b00) == 1
        assert len(lock_keys_b21) == 0  # stale lock was ignored
        assert len(list(lock.store.list("locks"))) == 0  # stale lock was removed from store

    def test_release_stale_lock(self, lockstore):
        # even if our own lock became stale (e.g. machine suspended while doing a backup or
        # a long time without repository access), release() must find and remove it instead
        # of removing it as stale and then raising NotLocked, see #9883.
        lock = Lock(lockstore, exclusive=True, id=ID1, stale=2)
        lock.acquire()
        time.sleep(2.1)  # lock is now older than the stale timeout
        lock.release()  # must not raise NotLocked
        assert len(list(lockstore.list("locks"))) == 0

    def test_refresh_stale_lock(self, lockstore):
        # if our own lock became stale, but no other client killed it (it is still present),
        # refresh() must renew it, so the operation can continue safely, see #9883.
        lock = Lock(lockstore, exclusive=True, id=ID1, stale=2)
        lock.acquire()
        old_keys = set(lock._get_locks())
        time.sleep(2.1)  # lock is now older than the stale timeout
        lock.refresh()  # must not raise LockTimeout, must renew the lock
        new_keys = set(lock._get_locks())
        assert len(old_keys) == len(new_keys) == 1
        assert old_keys != new_keys  # refresh done, new lock has different key
        lock.release()

    def test_refresh_killed_lock(self, lockstore):
        # if our own lock is gone (another client considered it stale and killed it),
        # there is no safe way to continue, refresh() must raise LockTimeout.
        lock = Lock(lockstore, exclusive=True, id=ID1, stale=2)
        lock.acquire()
        time.sleep(1.1)  # older than refresh_td (stale // 2), so refresh() checks the store
        Lock(lockstore, exclusive=True, id=ID2, stale=2).break_lock()  # kill it, like another client would
        with pytest.raises(LockTimeout):
            lock.refresh()

    def test_refresh_killed_lock_race(self, lockstore, monkeypatch):
        # if our own lock gets killed by another client *between* refresh() listing it and
        # deleting it (the other client considered it stale and might have acquired its own
        # lock already), refresh() must raise LockTimeout and must not leave its just-created
        # new lock behind (it would needlessly block other clients until it expired as stale).
        lock = Lock(lockstore, exclusive=True, id=ID1, stale=2)
        lock.acquire()
        old_key = lock.my_lock_key
        time.sleep(1.1)  # older than refresh_td (stale // 2), so refresh() will renew the lock

        orig_delete = lockstore.delete

        def delete(name, *args, **kwargs):
            if name == f"locks/{old_key}":
                orig_delete(name)  # another client killed our old lock just before we delete it,
                # so our own deletion attempt below finds it already gone and raises ObjectNotFound:
            return orig_delete(name, *args, **kwargs)

        monkeypatch.setattr(lockstore, "delete", delete)
        with pytest.raises(LockTimeout):
            lock.refresh()
        assert len(list(lockstore.list("locks"))) == 0  # no new lock left behind

    def test_migrate_lock(self, lockstore):
        old_id, new_id = ID1, ID2
        assert old_id[1] != new_id[1]  # different PIDs (like when doing daemonize())
        lock = Lock(lockstore, id=old_id).acquire()
        old_locks = lock._find_locks(only_mine=True)
        assert lock.id == old_id  # lock is for old id / PID
        lock.migrate_lock(old_id, new_id)  # fix the lock
        assert lock.id == new_id  # lock corresponds to the new id / PID
        new_locks = lock._find_locks(only_mine=True)
        assert old_locks != new_locks
        assert len(old_locks) == len(new_locks) == 1
        assert old_locks[0]["hostid"] == old_id[0]
        assert new_locks[0]["hostid"] == new_id[0]



from collections.abc import Iterator, MutableMapping
from typing import IO, Any, NamedTuple

PATH_OR_FILE = str | IO

class HTProxyMixin(MutableMapping): ...

class ChunkIndexEntry(NamedTuple):
    flags: int
    size: int  # plaintext chunk size
    pack_id: bytes
    obj_offset: int
    obj_size: int

CIE = tuple[int, int, bytes, int, int] | type[ChunkIndexEntry]

class ChunkIndexEntryFormatT(NamedTuple):
    flags: str
    size: str
    pack_id: str
    obj_offset: str
    obj_size: str

ChunkIndexEntryFormat: ChunkIndexEntryFormatT

class ChunkIndex:
    F_NONE: int
    F_USED: int
    F_COMPRESS: int
    F_NEW: int
    M_USER: int
    M_SYSTEM: int
    def add(self, key: bytes, size: int) -> None: ...
    def update_pack_info(self, pack_results: list | None) -> None: ...
    def iteritems(self, *, only_new: bool = ...) -> Iterator: ...
    def clear_new(self) -> None: ...
    def __contains__(self, key: bytes) -> bool: ...
    def __getitem__(self, key: bytes) -> type[ChunkIndexEntry]: ...
    def __setitem__(self, key: bytes, value: CIE) -> None: ...

class FuseVersionsIndexEntry(NamedTuple):
    version: int
    hash: bytes

class FuseVersionsIndex:
    def __contains__(self, key: bytes) -> bool: ...
    def __getitem__(self, key: bytes) -> Any: ...
    def __setitem__(self, key: bytes, value: Any) -> None: ...



import errno
import functools
import inspect
import logging
import os
import queue
import select
import shlex
import sys
import textwrap
import time
import traceback
from subprocess import Popen, PIPE

import borg.logger
from .. import __version__
from ..constants import *  # NOQA
from ..helpers import Error, ErrorWithTraceback, IntegrityError, PathNotAllowed
from ..helpers import bin_to_hex
from ..helpers import get_limited_unpacker
from ..helpers import replace_placeholders
from ..helpers import sysinfo
from ..helpers import format_file_size
from ..helpers import prepare_subprocess_env, ignore_sigint
from ..fslocking import LockTimeout, NotLocked, NotMyLock, LockFailed
from ..logger import create_logger, borg_serve_log_queue
from ..helpers import msgpack
from .repository import LegacyRepository
from ..repository import Repository, StoreObjectNotFound
from ..version import parse_version, format_version
from ..helpers.datastruct import EfficientCollectionQueue
from ..platform import is_win32

logger = create_logger(__name__)

BORG_VERSION = parse_version(__version__)
MSGID, MSG, ARGS, RESULT, LOG = "i", "m", "a", "r", "l"

MAX_INFLIGHT = 100

RATELIMIT_PERIOD = 0.1


class ConnectionClosed(Error):
    """Connection closed by remote host."""

    exit_mcode = 80


class ConnectionClosedWithHint(ConnectionClosed):
    """Connection closed by remote host. {}"""

    exit_mcode = 81


class InvalidRPCMethod(Error):
    """RPC method {} is not valid."""

    exit_mcode = 82


class UnexpectedRPCDataFormatFromClient(Error):
    """Borg {}: Got unexpected RPC data format from client."""

    exit_mcode = 85


class UnexpectedRPCDataFormatFromServer(Error):
    """Got unexpected RPC data format from server:\n{}"""

    exit_mcode = 86

    def __init__(self, data):
        try:
            data = data.decode()[:128]
        except UnicodeDecodeError:
            data = data[:128]
            data = ["%02X" % byte for byte in data]
            data = textwrap.fill(" ".join(data), 16 * 3)
        super().__init__(data)


class ConnectionBrokenWithHint(Error):
    """Connection to remote host is broken. {}"""

    exit_mcode = 87


# Protocol compatibility:
# In general the server is responsible for rejecting too old clients and the client is responsible for rejecting
# too old servers. This ensures that the knowledge what is compatible is always held by the newer component.
#
# For the client the return of the negotiate method is a dict which includes the server version.
#
# All method calls on the remote repository object must be allowlisted in RepositoryServer.rpc_methods and have api
# stubs in LegacyRemoteRepository. The @api decorator on these stubs is used to set server version requirements.
#
# Method parameters are identified only by name and never by position. Unknown parameters are ignored by the server.
# If a new parameter is important and may not be ignored, on the client a parameter specific version requirement needs
# to be added.
# When parameters are removed, they need to be preserved as defaulted parameters on the client stubs so that older
# servers still get compatible input.


class SleepingBandwidthLimiter:
    def __init__(self, limit):
        if limit:
            self.ratelimit = int(limit * RATELIMIT_PERIOD)
            self.ratelimit_last = time.monotonic()
            self.ratelimit_quota = self.ratelimit
        else:
            self.ratelimit = None

    def write(self, fd, to_send):
        if self.ratelimit:
            now = time.monotonic()
            if self.ratelimit_last + RATELIMIT_PERIOD <= now:
                self.ratelimit_quota += self.ratelimit
                if self.ratelimit_quota > 2 * self.ratelimit:
                    self.ratelimit_quota = 2 * self.ratelimit
                self.ratelimit_last = now
            if self.ratelimit_quota == 0:
                tosleep = self.ratelimit_last + RATELIMIT_PERIOD - now
                time.sleep(tosleep)
                self.ratelimit_quota += self.ratelimit
                self.ratelimit_last = time.monotonic()
            if len(to_send) > self.ratelimit_quota:
                to_send = to_send[: self.ratelimit_quota]
        try:
            written = os.write(fd, to_send)
        except BrokenPipeError:
            raise ConnectionBrokenWithHint("Broken Pipe") from None
        if self.ratelimit:
            self.ratelimit_quota -= written
        return written


def api(*, since, **kwargs_decorator):
    """Check version requirements and use self.call to do the remote method call.

    <since> specifies the version in which borg introduced this method.
    Calling this method when connected to an older version will fail without transmitting anything to the server.

    Further kwargs can be used to encode version specific restrictions:

    <previously> is the value resulting in the behaviour before introducing the new parameter.
    If a previous hardcoded behaviour is parameterized in a version, this allows calls that use the previously
    hardcoded behaviour to pass through and generates an error if another behaviour is requested by the client.
    E.g. when 'append_only' was introduced in 1.0.7 the previous behaviour was what now is append_only=False.
    Thus @api(..., append_only={'since': parse_version('1.0.7'), 'previously': False}) allows calls
    with append_only=False for all version but rejects calls using append_only=True on versions older than 1.0.7.

    <dontcare> is a flag to set the behaviour if an old version is called the new way.
    If set to True, the method is called without the (not yet supported) parameter (this should be done if that is the
    more desirable behaviour). If False, an exception is generated.
    E.g. before 'threshold' was introduced in 1.2.0a8, a hardcoded threshold of 0.1 was used in commit().
    """

    def decorator(f):
        @functools.wraps(f)
        def do_rpc(self, *args, **kwargs):
            sig = inspect.signature(f)
            bound_args = sig.bind(self, *args, **kwargs)
            named = {}  # Arguments for the remote process
            extra = {}  # Arguments for the local process
            for name, param in sig.parameters.items():
                if name == "self":
                    continue
                if name in bound_args.arguments:
                    if name == "wait":
                        extra[name] = bound_args.arguments[name]
                    else:
                        named[name] = bound_args.arguments[name]
                else:
                    if param.default is not param.empty:
                        named[name] = param.default

            if self.server_version < since:
                raise self.RPCServerOutdated(f.__name__, format_version(since))

            for name, restriction in kwargs_decorator.items():
                if restriction["since"] <= self.server_version:
                    continue
                if "previously" in restriction and named[name] == restriction["previously"]:
                    continue
                if restriction.get("dontcare", False):
                    continue

                raise self.RPCServerOutdated(
                    f"{f.__name__} {name}={named[name]!s}", format_version(restriction["since"])
                )

            return self.call(f.__name__, named, **extra)

        return do_rpc

    return decorator


class LegacyRemoteRepository:
    extra_test_args = []  # type: ignore

    class RPCError(Exception):
        def __init__(self, unpacked):
            # unpacked has keys: 'exception_args', 'exception_full', 'exception_short', 'sysinfo'
            self.unpacked = unpacked

        def get_message(self):
            return "\n".join(self.unpacked["exception_short"])

        @property
        def traceback(self):
            return self.unpacked.get("exception_trace", True)

        @property
        def exception_class(self):
            return self.unpacked["exception_class"]

        @property
        def exception_full(self):
            return "\n".join(self.unpacked["exception_full"])

        @property
        def sysinfo(self):
            return self.unpacked["sysinfo"]

    class RPCServerOutdated(Error):
        """Borg server is too old for {}. Required version {}"""

        exit_mcode = 84

        @property
        def method(self):
            return self.args[0]

        @property
        def required_version(self):
            return self.args[1]

    def __init__(self, location, create=False, exclusive=False, lock_wait=None, lock=True, args=None):
        self.location = self._location = location
        self.msgid = 0
        self.rx_bytes = 0
        self.tx_bytes = 0
        self.to_send = EfficientCollectionQueue(1024 * 1024, bytes)
        self.stdin_fd = self.stdout_fd = self.stderr_fd = None
        self.stderr_received = b""  # incomplete stderr line bytes received (no \n yet)
        self.ignore_responses = set()
        self.responses = {}
        self.async_responses = {}
        self.shutdown_time = None
        self.ratelimit = SleepingBandwidthLimiter(args.upload_ratelimit * 1024 if args and args.upload_ratelimit else 0)
        self.upload_buffer_size_limit = args.upload_buffer * 1024 * 1024 if args and args.upload_buffer else 0
        self.unpacker = get_limited_unpacker("client")
        self.server_version = None  # we update this after server sends its version
        self.p = None
        self._args = args
        if self.location.proto == "ssh":
            testing = location.host == "__testsuite__"
            # when testing, we invoke and talk to a borg process directly (no ssh).
            # when not testing, we invoke the system-installed ssh binary to talk to a remote borg.
            env = prepare_subprocess_env(system=not testing)
            borg_cmd = self.borg_cmd(args, testing)
            if not testing:
                borg_cmd = self.ssh_cmd(location) + borg_cmd
            logger.debug("SSH command line: %s", borg_cmd)
            # we do not want the ssh getting killed by Ctrl-C/SIGINT because it is needed for clean shutdown of borg.
            self.p = Popen(
                borg_cmd,
                bufsize=0,
                stdin=PIPE,
                stdout=PIPE,
                stderr=PIPE,
                env=env,
                preexec_fn=None if is_win32 else ignore_sigint,
            )  # nosec B603
            self.stdin_fd = self.p.stdin.fileno()
            self.stdout_fd = self.p.stdout.fileno()
            self.stderr_fd = self.p.stderr.fileno()
            self.r_fds = [self.stdout_fd, self.stderr_fd]
            self.x_fds = [self.stdin_fd, self.stdout_fd, self.stderr_fd]
        else:
            raise Error(f"Unsupported protocol {location.proto}")

        os.set_blocking(self.stdin_fd, False)
        assert not os.get_blocking(self.stdin_fd)
        os.set_blocking(self.stdout_fd, False)
        assert not os.get_blocking(self.stdout_fd)
        if self.stderr_fd is not None:
            os.set_blocking(self.stderr_fd, False)
            assert not os.get_blocking(self.stderr_fd)

        try:
            try:
                version = self.call("negotiate", {"client_data": {"client_version": BORG_VERSION}})
            except ConnectionClosed:
                raise ConnectionClosedWithHint("Is borg working on the server?") from None
            if isinstance(version, dict):
                self.server_version = version["server_version"]
            else:
                raise Exception("Server insisted on using unsupported protocol version %s" % version)

            self.id = self.open(
                path=self.location.path,
                create=create,
                lock_wait=lock_wait,
                lock=lock,
                exclusive=exclusive,
                v1_legacy=True,  # make remote use LegacyRepository
            )
            info = self.info()
            self.version = info["version"]

        except Exception:
            self.close()
            raise

    def __del__(self):
        if len(self.responses):
            logging.debug("still %d cached responses left in LegacyRemoteRepository" % (len(self.responses),))
        if self.p:
            self.close()
            assert False, "cleanup happened in LegacyRemoteRepository.__del__"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.location.canonical_path()}>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is not None:
                self.shutdown_time = time.monotonic() + 30
                self.rollback()
        finally:
            # in any case, we want to close the repo cleanly, even if the
            # rollback can not succeed (e.g. because the connection was
            # already closed) and raised another exception:
            logger.debug(
                "LegacyRemoteRepository: %s bytes sent, %s bytes received, %d messages sent",
                format_file_size(self.tx_bytes),
                format_file_size(self.rx_bytes),
                self.msgid,
            )
            self.close()

    @property
    def id_str(self):
        return bin_to_hex(self.id)

    def borg_cmd(self, args, testing):
        """return a borg serve command line"""
        # give some args/options to 'borg serve' process as they were given to us
        opts = []
        if args is not None:
            root_logger = logging.getLogger()
            if root_logger.isEnabledFor(logging.DEBUG):
                opts.append("--debug")
            elif root_logger.isEnabledFor(logging.INFO):
                opts.append("--info")
            elif root_logger.isEnabledFor(logging.WARNING):
                pass  # warning is default
            elif root_logger.isEnabledFor(logging.ERROR):
                opts.append("--error")
            elif root_logger.isEnabledFor(logging.CRITICAL):
                opts.append("--critical")
            else:
                raise ValueError("log level missing, fix this code")

            # Tell the remote server about debug topics it may need to consider.
            # Note that debug topics are usable for "spew" or "trace" logs which would
            # be too plentiful to transfer for normal use, so the server doesn't send
            # them unless explicitly enabled.
            #
            # Needless to say, if you do --debug-topic=repository.compaction, for example,
            # with a 1.0.x server it won't work, because the server does not recognize the
            # option.
            #
            # This is not considered a problem, since this is a debugging feature that
            # should not be used for regular use.
            for topic in args.debug_topics:
                if "." not in topic:
                    topic = "borg.debug." + topic
                if "repository" in topic:
                    opts.append("--debug-topic=%s" % topic)
        env_vars = []
        if testing:
            return env_vars + [sys.executable, "-m", "borg", "serve"] + opts + self.extra_test_args
        else:  # pragma: no cover
            remote_path = args.remote_path or os.environ.get("BORG_REMOTE_PATH", "borg")
            remote_path = replace_placeholders(remote_path)
            return env_vars + [remote_path, "serve"] + opts

    def ssh_cmd(self, location):
        """return a ssh command line that can be prefixed to a borg command line"""
        rsh = self._args.rsh or os.environ.get("BORG_RSH", "ssh")
        args = shlex.split(rsh)
        if location.port:
            args += ["-p", str(location.port)]
        if location.user:
            args.append(f"{location.user}@{location.host}")
        else:
            args.append("%s" % location.host)
        return args

    def call(self, cmd, args, **kw):
        for resp in self.call_many(cmd, [args], **kw):
            return resp

    def call_many(self, cmd, calls, wait=True, async_wait=True):
        if not calls and cmd != "async_responses":
            return

        def send_buffer():
            if self.to_send:
                try:
                    written = self.ratelimit.write(self.stdin_fd, self.to_send.peek_front())
                    self.tx_bytes += written
                    self.to_send.pop_front(written)
                except OSError as e:
                    # io.write might raise EAGAIN even though select indicates
                    # that the fd should be writable.
                    # EWOULDBLOCK is added for defensive programming sake.
                    if e.errno not in [errno.EAGAIN, errno.EWOULDBLOCK]:
                        raise

        def handle_error(unpacked):
            if "exception_class" not in unpacked:
                return

            error = unpacked["exception_class"]
            args = unpacked["exception_args"]

            if error == "Error":
                raise Error(args[0])
            elif error == "ErrorWithTraceback":
                raise ErrorWithTraceback(args[0])
            elif error == "DoesNotExist":
                raise LegacyRepository.DoesNotExist(self.location.processed)
            elif error == "AlreadyExists":
                raise LegacyRepository.AlreadyExists(self.location.processed)
            elif error == "CheckNeeded":
                raise LegacyRepository.CheckNeeded(self.location.processed)
            elif error == "IntegrityError":
                raise IntegrityError(args[0])
            elif error == "PathNotAllowed":
                raise PathNotAllowed(args[0])
            elif error == "PathPermissionDenied":
                raise LegacyRepository.PathPermissionDenied(args[0])
            elif error == "ParentPathDoesNotExist":
                raise LegacyRepository.ParentPathDoesNotExist(args[0])
            elif error == "ObjectNotFound":
                raise LegacyRepository.ObjectNotFound(args[0], self.location.processed)
            elif error == "InvalidRPCMethod":
                raise InvalidRPCMethod(args[0])
            elif error == "LockTimeout":
                raise LockTimeout(args[0])
            elif error == "LockFailed":
                raise LockFailed(args[0], args[1])
            elif error == "NotLocked":
                raise NotLocked(args[0])
            elif error == "NotMyLock":
                raise NotMyLock(args[0])
            else:
                raise self.RPCError(unpacked)

        calls = list(calls)
        waiting_for = []
        maximum_to_send = 0 if wait else self.upload_buffer_size_limit
        send_buffer()  # Try to send data, as some cases (async_response) will never try to send data otherwise.
        try:
            while wait or calls:
                if self.shutdown_time and time.monotonic() > self.shutdown_time:
                    # we are shutting this LegacyRemoteRepository down already, make sure we do not waste
                    # a lot of time in case a lot of async stuff is coming in or remote is gone or slow.
                    logger.debug(
                        "shutdown_time reached, shutting down with %d waiting_for and %d async_responses.",
                        len(waiting_for),
                        len(self.async_responses),
                    )
                    return
                while waiting_for:
                    try:
                        unpacked = self.responses.pop(waiting_for[0])
                        waiting_for.pop(0)
                        handle_error(unpacked)
                        yield unpacked[RESULT]
                        if not waiting_for and not calls:
                            return
                    except KeyError:
                        break
                if cmd == "async_responses":
                    while True:
                        try:
                            msgid, unpacked = self.async_responses.popitem()
                        except KeyError:
                            # there is nothing left what we already have received
                            if async_wait and self.ignore_responses:
                                # but do not return if we shall wait and there is something left to wait for:
                                break
                            else:
                                return
                        else:
                            handle_error(unpacked)
                            yield unpacked[RESULT]
                if self.to_send or (calls and len(waiting_for) < MAX_INFLIGHT):
                    w_fds = [self.stdin_fd]
                else:
                    w_fds = []
                r, w, x = select.select(self.r_fds, w_fds, self.x_fds, 1)
                if x:
                    raise Exception("FD exception occurred")
                for fd in r:
                    if fd is self.stdout_fd:
                        data = os.read(fd, BUFSIZE)
                        if not data:
                            raise ConnectionClosed()
                        self.rx_bytes += len(data)
                        self.unpacker.feed(data)
                        for unpacked in self.unpacker:
                            if not isinstance(unpacked, dict):
                                raise UnexpectedRPCDataFormatFromServer(data)

                            lr_dict = unpacked.get(LOG)
                            if lr_dict is not None:
                                # Re-emit remote log messages locally.
                                _logger = logging.getLogger(lr_dict["name"])
                                if _logger.isEnabledFor(lr_dict["level"]):
                                    _logger.handle(logging.LogRecord(**lr_dict))
                                continue

                            msgid = unpacked[MSGID]
                            if msgid in self.ignore_responses:
                                self.ignore_responses.remove(msgid)
                                # async methods never return values, but may raise exceptions.
                                if "exception_class" in unpacked:
                                    self.async_responses[msgid] = unpacked
                                else:
                                    # we currently do not have async result values except "None",
                                    # so we do not add them into async_responses.
                                    if unpacked[RESULT] is not None:
                                        self.async_responses[msgid] = unpacked
                            else:
                                self.responses[msgid] = unpacked
                    elif fd is self.stderr_fd:
                        data = os.read(fd, 32768)
                        if not data:
                            raise ConnectionClosed()
                        self.rx_bytes += len(data)
                        # deal with incomplete lines (may appear due to block buffering)
                        if self.stderr_received:
                            data = self.stderr_received + data
                            self.stderr_received = b""
                        lines = data.splitlines(keepends=True)
                        if lines and not lines[-1].endswith((b"\r", b"\n")):
                            self.stderr_received = lines.pop()
                        # now we have complete lines in <lines> and any partial line in self.stderr_received.
                        _logger = logging.getLogger()
                        for line in lines:
                            # borg serve (remote/server side) should not emit stuff on stderr,
                            # but e.g. the ssh process (local/client side) might output errors there.
                            assert line.endswith((b"\r", b"\n"))
                            # something came in on stderr, log it to not lose it.
                            # decode late, avoid partial utf-8 sequences.
                            _logger.warning("stderr: " + line.decode().strip())
                if w:
                    while (len(self.to_send) <= maximum_to_send) and calls and len(waiting_for) < MAX_INFLIGHT:
                        args = calls.pop(0)
                        self.msgid += 1
                        waiting_for.append(self.msgid)
                        self.to_send.push_back(msgpack.packb({MSGID: self.msgid, MSG: cmd, ARGS: args}))

                    send_buffer()
        finally:
            self.ignore_responses |= set(waiting_for)  # we lose order here

    @api(since=parse_version("1.0.0"), v1_legacy={"since": parse_version("2.0.0b21"), "previously": True})
    def open(self, path, create=False, lock_wait=None, lock=True, exclusive=False, v1_legacy=False):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("2.0.0a3"))
    def info(self):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"), max_duration={"since": parse_version("1.2.0a4"), "previously": 0})
    def check(self, repair=False, max_duration=0):
        """actual remoting is done via self.call in the @api decorator"""

    @api(
        since=parse_version("1.0.0"),
        compact={"since": parse_version("1.2.0a0"), "previously": True, "dontcare": True},
        threshold={"since": parse_version("1.2.0a8"), "previously": 0.1, "dontcare": True},
    )
    def commit(self, compact=True, threshold=0.1):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def rollback(self):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def destroy(self):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def __len__(self):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def list(self, limit=None, marker=None):
        """actual remoting is done via self.call in the @api decorator"""

    def get(self, id, read_data=True, raise_missing=True):
        for resp in self.get_many([id], read_data=read_data, raise_missing=raise_missing):
            return resp

    def get_many(self, ids, read_data=True, raise_missing=True):
        # note: legacy remote protocol does not support raise_missing parameter, so we ignore it here
        yield from self.call_many("get", [{"id": id, "read_data": read_data} for id in ids])

    @api(since=parse_version("1.0.0"))
    def put(self, id, data, wait=True):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def delete(self, id, wait=True):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def save_key(self, keydata):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def load_key(self):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("1.0.0"))
    def break_lock(self):
        """actual remoting is done via self.call in the @api decorator"""

    def close(self):
        if self.p:
            self.call("close", {}, wait=True)
            self.p.stdin.close()
            self.p.stdout.close()
            self.p.wait()
            self.p = None

    def async_response(self, wait=True):
        for resp in self.call_many("async_responses", calls=[], wait=True, async_wait=wait):
            return resp

    @api(since=parse_version("2.0.0b8"))
    def get_manifest(self):
        """actual remoting is done via self.call in the @api decorator"""

    @api(since=parse_version("2.0.0b8"))
    def put_manifest(self, data):
        """actual remoting is done via self.call in the @api decorator"""


# borg serve: borg only serves legacy (borg 1.x / v1) repositories over ssh:// now (current
# repositories use rest://). The legacy client above (LegacyRemoteRepository) spawns "borg serve"
# on the remote host; this server keeps the legacy RPC method allowlist and opens LegacyRepository.


class RepositoryServer:  # pragma: no cover
    _legacy_rpc_methods = (  # LegacyRepository
        "__len__",
        "check",
        "commit",
        "delete",
        "destroy",
        "get",
        "list",
        "negotiate",
        "open",
        "close",
        "info",
        "put",
        "rollback",
        "save_key",
        "load_key",
        "break_lock",
        "inject_exception",
        "get_manifest",  # borg2 LegacyRepository has this
    )

    def __init__(self, restrict_to_paths, restrict_to_repositories):
        self.repository = None
        self.RepoCls = None
        self.rpc_methods = ("open", "close", "negotiate")
        self.restrict_to_paths = restrict_to_paths
        self.restrict_to_repositories = restrict_to_repositories
        # note: legacy (borg 1.x / v1) repositories have no permission system, so borg serve
        # does not accept/forward permissions here (the --permissions option only applies to
        # the non-legacy "borg serve --rest" path).
        self.client_version = None  # we update this after client sends version information

    def filter_args(self, f, kwargs):
        """Remove unknown named parameters from call, because client did (implicitly) say it's ok."""
        known = set(inspect.signature(f).parameters)
        return {name: kwargs[name] for name in kwargs if name in known}

    def send_queued_log(self):
        while True:
            try:
                # lr_dict contents see BorgQueueHandler
                lr_dict = borg_serve_log_queue.get_nowait()
            except queue.Empty:
                break
            else:
                msg = msgpack.packb({LOG: lr_dict})
                os.write(self.stdout_fd, msg)

    def serve(self):
        def inner_serve():
            os.set_blocking(self.stdin_fd, False)
            assert not os.get_blocking(self.stdin_fd)
            os.set_blocking(self.stdout_fd, True)
            assert os.get_blocking(self.stdout_fd)

            unpacker = get_limited_unpacker("server")
            shutdown_serve = False
            while True:
                # before processing any new RPCs, send out all pending log output
                self.send_queued_log()

                if shutdown_serve:
                    # shutdown wanted! get out of here after sending all log output.
                    assert self.repository is None
                    return

                # process new RPCs
                r, w, es = select.select([self.stdin_fd], [], [], 10)
                if r:
                    data = os.read(self.stdin_fd, BUFSIZE)
                    if not data:
                        shutdown_serve = True
                        continue
                    unpacker.feed(data)
                    for unpacked in unpacker:
                        if isinstance(unpacked, dict):
                            msgid = unpacked[MSGID]
                            method = unpacked[MSG]
                            args = unpacked[ARGS]
                        else:
                            if self.repository is not None:
                                self.repository.close()
                            raise UnexpectedRPCDataFormatFromClient(__version__)
                        try:
                            # logger.debug(f"{type(self)} method: {type(self.repository)}.{method}")
                            if method not in self.rpc_methods:
                                raise InvalidRPCMethod(method)
                            try:
                                f = getattr(self, method)
                            except AttributeError:
                                f = getattr(self.repository, method)
                            args = self.filter_args(f, args)
                            res = f(**args)
                        except BaseException as e:
                            # These exceptions are reconstructed on the client end in
                            # LegacyRemoteRepository.call_many(), and will be handled just like locally raised
                            # exceptions. Suppress the remote traceback for these, except ErrorWithTraceback,
                            # which should always display a traceback.
                            reconstructed_exceptions = (
                                Repository.InvalidRepository,
                                Repository.InvalidRepositoryConfig,
                                Repository.DoesNotExist,
                                Repository.AlreadyExists,
                                Repository.PathAlreadyExists,
                                PathNotAllowed,
                                Repository.InsufficientFreeSpaceError,
                            )
                            # logger.exception(e)
                            ex_short = traceback.format_exception_only(e.__class__, e)
                            ex_full = traceback.format_exception(*sys.exc_info())
                            ex_trace = True
                            if isinstance(e, Error):
                                ex_short = [e.get_message()]
                                ex_trace = e.traceback
                            if not isinstance(e, reconstructed_exceptions):
                                logging.debug("\n".join(ex_full))

                            sys_info = sysinfo()
                            # StoreObjectNotFound and Repository.ObjectNotFound both have
                            # __name__ == "ObjectNotFound", so we need to distinguish them
                            # explicitly for correct client-side reconstruction.
                            exc_cls_name = (
                                "StoreObjectNotFound" if isinstance(e, StoreObjectNotFound) else e.__class__.__name__
                            )
                            try:
                                msg = msgpack.packb(
                                    {
                                        MSGID: msgid,
                                        "exception_class": exc_cls_name,
                                        "exception_args": e.args,
                                        "exception_full": ex_full,
                                        "exception_short": ex_short,
                                        "exception_trace": ex_trace,
                                        "sysinfo": sys_info,
                                    }
                                )
                            except TypeError:
                                msg = msgpack.packb(
                                    {
                                        MSGID: msgid,
                                        "exception_class": exc_cls_name,
                                        "exception_args": [
                                            x if isinstance(x, (str, bytes, int)) else None for x in e.args
                                        ],
                                        "exception_full": ex_full,
                                        "exception_short": ex_short,
                                        "exception_trace": ex_trace,
                                        "sysinfo": sys_info,
                                    }
                                )
                            os.write(self.stdout_fd, msg)
                        else:
                            os.write(self.stdout_fd, msgpack.packb({MSGID: msgid, RESULT: res}))
                if es:
                    shutdown_serve = True
                    continue

        # server for one ssh:// connection
        self.stdin_fd = sys.stdin.fileno()
        self.stdout_fd = sys.stdout.fileno()
        inner_serve()

    def negotiate(self, client_data):
        if isinstance(client_data, dict):
            self.client_version = client_data["client_version"]
        else:
            self.client_version = BORG_VERSION  # seems to be newer than current version (no known old format)

        # not a known old format, send newest negotiate this version knows
        return {"server_version": BORG_VERSION}

    def _resolve_path(self, path):
        if isinstance(path, bytes):
            path = os.fsdecode(path)
        path = os.path.realpath(path)
        return path

    def open(self, path, create=False, lock_wait=None, lock=True, exclusive=None, v1_legacy=False):
        # borg only serves legacy (v1) repositories now; current repositories are accessed via rest://.
        self.RepoCls = LegacyRepository
        self.rpc_methods = self._legacy_rpc_methods
        logging.debug("Resolving repository path %r", path)
        path = self._resolve_path(path)
        logging.debug("Resolved repository path to %r", path)
        path_with_sep = os.path.join(path, "")  # make sure there is a trailing slash (os.sep)
        if self.restrict_to_paths:
            # if --restrict-to-path P is given, we make sure that we only operate in/below path P.
            # for the prefix check, it is important that the compared paths both have trailing slashes,
            # so that a path /foobar will NOT be accepted with --restrict-to-path /foo option.
            for restrict_to_path in self.restrict_to_paths:
                restrict_to_path_with_sep = os.path.join(os.path.realpath(restrict_to_path), "")  # trailing slash
                if path_with_sep.startswith(restrict_to_path_with_sep):
                    break
            else:
                raise PathNotAllowed(path)
        if self.restrict_to_repositories:
            for restrict_to_repository in self.restrict_to_repositories:
                restrict_to_repository_with_sep = os.path.join(os.path.realpath(restrict_to_repository), "")
                if restrict_to_repository_with_sep == path_with_sep:
                    break
            else:
                raise PathNotAllowed(path)
        kwargs = dict(lock_wait=lock_wait, lock=lock, exclusive=exclusive, send_log_cb=self.send_queued_log)
        self.repository = self.RepoCls(path, create, **kwargs)
        self.repository.__enter__()  # clean exit handled by serve() method
        return self.repository.id

    def close(self):
        if self.repository is not None:
            self.repository.__exit__(None, None, None)
            self.repository = None
        borg.logger.flush_logging()
        self.send_queued_log()

    def inject_exception(self, kind):
        s1 = "test string"
        s2 = "test string2"
        if kind == "DoesNotExist":
            raise self.RepoCls.DoesNotExist(s1)
        elif kind == "AlreadyExists":
            raise self.RepoCls.AlreadyExists(s1)
        elif kind == "CheckNeeded":
            raise self.RepoCls.CheckNeeded(s1)
        elif kind == "IntegrityError":
            raise IntegrityError(s1)
        elif kind == "PathNotAllowed":
            raise PathNotAllowed("foo")
        elif kind == "ObjectNotFound":
            raise self.RepoCls.ObjectNotFound(s1, s2)
        elif kind == "StoreObjectNotFound":
            raise StoreObjectNotFound(s1)
        elif kind == "InvalidRPCMethod":
            raise InvalidRPCMethod(s1)
        elif kind == "divide":
            0 // 0



# cython: language_level=3



import cython
import time

from cpython.bytes cimport PyBytes_AsString
from libc.stdint cimport uint8_t, uint64_t
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, memmove, memset

from ..crypto.low_level import CSPRNG

from ..constants import CH_DATA, CH_ALLOC, CH_HOLE, zeros
from .reader import FileReader, Chunk

# Cyclic polynomial / buzhash
#
# https://en.wikipedia.org/wiki/Rolling_hash
#
# http://www.serve.net/buz/Notes.1st.year/HTML/C6/rand.012.html (by "BUZ", the inventor)
#
# http://www.dcs.gla.ac.uk/~hamer/cakes-talk.pdf (see buzhash slide)
#
# Some properties of buzhash / of this implementation:
#
# (1) the hash is designed for inputs <= 64 bytes, but the chunker uses it on a 4095 byte window;
#     any repeating bytes at distance 64 within those 4095 bytes can cause cancellation within
#     the hash function, e.g. in "X <any 63 bytes> X", the last X would cancel out the influence
#     of the first X on the hash value.

# This seems to be the most reliable way to inline this code, using a C preprocessor macro:
cdef extern from *:
   """
   #define BARREL_SHIFT64(v, shift) (((v) << (shift)) | ((v) >> (((64 - (shift)) & 0x3f))))
   """
   uint64_t BARREL_SHIFT64(uint64_t v, uint64_t shift)


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)  # Deactivate negative indexing.
cdef uint64_t* buzhash64_init_table(bytes key) except NULL:
    """
    Generate a balanced pseudo-random table deterministically from a 256-bit key.
    Balanced means that for each bit position 0..63, exactly 50% of the table values have the bit set to 1.
    """
    # Create deterministic random number generator
    rng = CSPRNG(key)

    cdef int i, j, bit_pos
    cdef uint64_t* table = <uint64_t*>malloc(2048)  # 256 * sizeof(uint64_t)
    if table == NULL:
        raise MemoryError("Failed to allocate buzhash64 table")

    # Initialize all values to 0
    for i in range(256):
        table[i] = 0

    # For each bit position, deterministically assign exactly 128 positions to have that bit set
    for bit_pos in range(64):
        # Create a list of indices and shuffle deterministically
        indices = list(range(256))
        rng.shuffle(indices)

        # Set the bit at bit_pos for the first 128 shuffled indices
        for i in range(128):
            j = indices[i]
            table[j] |= (1ULL << bit_pos)

    return table


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)  # Deactivate negative indexing.
@cython.cdivision(True)  # Use C division/modulo semantics for integer division.
cdef uint64_t _buzhash64(const unsigned char* data, size_t len, const uint64_t* h):
    """Calculate the buzhash of the given data."""
    cdef uint64_t i
    cdef uint64_t sum = 0, imod
    if len == 0:
        return 0
    for i in range(len - 1, 0, -1):
        imod = i & 0x3f
        sum ^= BARREL_SHIFT64(h[data[0]], imod)
        data += 1
    return sum ^ h[data[0]]


@cython.boundscheck(False)  # Deactivate bounds checking
@cython.wraparound(False)  # Deactivate negative indexing.
@cython.cdivision(True)  # Use C division/modulo semantics for integer division.
cdef uint64_t _buzhash64_update(uint64_t sum, unsigned char remove, unsigned char add, size_t len, const uint64_t* h):
    """Update the buzhash with a new byte."""
    cdef uint64_t lenmod = len & 0x3f
    return BARREL_SHIFT64(sum, 1) ^ BARREL_SHIFT64(h[remove], lenmod) ^ h[add]


cdef class ChunkerBuzHash64:
    """
    Content-Defined Chunker, variable chunk sizes.

    This chunker makes quite some effort to cut mostly chunks of the same-content, even if
    the content moves to a different offset inside the file. It uses the buzhash
    rolling-hash algorithm to identify the chunk cutting places by looking at the
    content inside the moving window and computing the rolling hash value over the
    window contents. If the last n bits of the rolling hash are 0, a chunk is cut.
    Additionally it obeys some more criteria, like a minimum and maximum chunk size.
    It also uses a per-repo random seed to avoid some chunk length fingerprinting attacks.
    """
    cdef uint64_t chunk_mask
    cdef uint64_t mask_s, mask_l  # normalized chunking: strict / loose masks
    cdef size_t normal_size       # chunk length at which we switch mask_s -> mask_l
    cdef int nc_level             # normalized chunking level (0 = disabled)
    cdef uint64_t* table
    cdef uint8_t* data
    cdef object _fd  # Python object for file descriptor
    cdef int fh
    cdef int done, eof
    cdef size_t min_size, buf_size, window_size, remaining, position, last
    cdef long long bytes_read, bytes_yielded  # off_t in C, using long long for compatibility
    cdef readonly float chunking_time
    cdef object file_reader  # FileReader instance
    cdef size_t reader_block_size
    cdef bint sparse

    def __cinit__(self, bytes key, int chunk_min_exp, int chunk_max_exp, int hash_mask_bits, int hash_window_size, int nc_level=0, size_t normal_size=0, bint sparse=False):
        self.table = NULL
        self.data = NULL
        min_size = 1 << chunk_min_exp
        max_size = 1 << chunk_max_exp
        assert max_size <= len(zeros)
        # see chunker_process, first while loop condition, first term must be able to get True:
        assert hash_window_size + min_size + 1 <= max_size, "too small max_size"

        self.window_size = hash_window_size
        self.chunk_mask = (1ULL << hash_mask_bits) - 1
        self.min_size = min_size
        # Normalized chunking (FastCDC-style): use a stricter mask (lower cut probability) until
        # the chunk reaches its expected/normal size, then a looser mask (higher cut probability).
        # This concentrates chunk sizes around the target and reduces chunk-size variance.
        # nc_level == 0 disables it, keeping behavior byte-identical to the single-mask chunker.
        assert nc_level >= 0
        assert hash_mask_bits - nc_level >= 1, "nc_level too large for hash_mask_bits"
        assert hash_mask_bits + nc_level <= 48, "nc_level too large for hash_mask_bits"
        self.nc_level = nc_level
        if nc_level:
            self.mask_s = (1ULL << (hash_mask_bits + nc_level)) - 1
            self.mask_l = (1ULL << (hash_mask_bits - nc_level)) - 1
            # normal_size is the chunk length at which we switch from the strict to the loose
            # mask; it dominates the mean chunk size. The default is the nominal target size
            # (1ULL << hash_mask_bits) minus the expected loose-phase tail (1ULL << (bits - nc_level)),
            # which lands the mean close to the target instead of overshooting it. Pass an
            # explicit normal_size to tune it further.
            self.normal_size = normal_size if normal_size else ((1ULL << hash_mask_bits) - (1ULL << (hash_mask_bits - nc_level)))
        else:
            self.mask_s = self.chunk_mask
            self.mask_l = self.chunk_mask
            self.normal_size = 0
        self.table = buzhash64_init_table(key)
        self.buf_size = max_size
        self.data = <uint8_t*>malloc(self.buf_size)
        if self.data == NULL:
            raise MemoryError("Failed to allocate chunker buffer")
        self.fh = -1
        self.done = 0
        self.eof = 0
        self.remaining = 0
        self.position = 0
        self.last = 0
        self.bytes_read = 0
        self.bytes_yielded = 0
        self._fd = None
        self.chunking_time = 0.0
        self.reader_block_size = 1024 * 1024
        self.sparse = sparse

    def __dealloc__(self):
        """Free the chunker's resources."""
        if self.table != NULL:
            free(self.table)
            self.table = NULL
        if self.data != NULL:
            free(self.data)
            self.data = NULL

    cdef int fill(self) except 0:
        """Fill the chunker's buffer with more data."""
        cdef ssize_t n
        cdef object chunk

        # Move remaining data to the beginning of the buffer
        memmove(self.data, self.data + self.last, self.position + self.remaining - self.last)
        self.position -= self.last
        self.last = 0
        n = self.buf_size - self.position - self.remaining

        if self.eof or n == 0:
            return 1

        # Use FileReader to read data
        chunk = self.file_reader.read(n)
        n = chunk.meta["size"]

        if n > 0:
            # Only copy data if it's not a hole
            if chunk.meta["allocation"] == CH_DATA:
                # Copy data from chunk to our buffer
                memcpy(self.data + self.position + self.remaining, <const unsigned char*>PyBytes_AsString(chunk.data), n)
            else:
                # For holes, fill with zeros using memset
                memset(self.data + self.position + self.remaining, 0, n)

            self.remaining += n
            self.bytes_read += n
        else:
            self.eof = 1

        return 1

    cdef object process(self) except *:
        """Process the chunker's buffer and return the next chunk."""
        cdef uint64_t sum, mask
        cdef uint64_t mask_s = self.mask_s, mask_l = self.mask_l
        cdef int nc_level = self.nc_level
        cdef size_t n, old_last, min_size = self.min_size, window_size = self.window_size
        cdef size_t normal_size = self.normal_size, normal_pos
        cdef uint8_t* p
        cdef uint8_t* stop_at
        cdef uint8_t* nc_stop
        cdef size_t did_bytes

        if self.done:
            if self.bytes_read == self.bytes_yielded:
                raise StopIteration
            else:
                raise Exception("chunkifier byte count mismatch")

        while self.remaining < min_size + window_size + 1 and not self.eof:  # see assert in Chunker init
            self.fill()

        # Here we either are at eof...
        if self.eof:
            self.done = 1
            if self.remaining:
                self.bytes_yielded += self.remaining
                # Return a memory view of the remaining data
                return memoryview((self.data + self.position)[:self.remaining])
            else:
                if self.bytes_read == self.bytes_yielded:
                    raise StopIteration
                else:
                    raise Exception("chunkifier byte count mismatch")

        # ... or we have at least min_size + window_size + 1 bytes remaining.
        # We do not want to "cut" a chunk smaller than min_size and the hash
        # window starts at the potential cutting place.
        self.position += min_size
        self.remaining -= min_size
        sum = _buzhash64(self.data + self.position, window_size, self.table)

        # Normalized chunking: pick the mask based on how far we are into the current chunk.
        # While below normal_size use the strict mask (lower cut probability), afterward the
        # loose mask (higher cut probability). The mask is re-evaluated at the top of every
        # iteration, so the transition is honored exactly at normal_pos. When nc is disabled,
        # mask_s == mask_l == chunk_mask and the normal_pos cap is not applied, so this reduces
        # to the original single-mask behavior.
        mask = mask_s
        normal_pos = 0
        while True:
            if nc_level:
                normal_pos = self.last + normal_size
                mask = mask_s if self.position < normal_pos else mask_l

            if not (self.remaining > window_size and (sum & mask) and not (self.eof and self.remaining <= window_size)):
                break

            p = self.data + self.position
            stop_at = p + self.remaining - window_size

            if nc_level and self.position < normal_pos:
                # do not scan past the strict->loose transition; re-evaluate the mask there
                nc_stop = self.data + normal_pos
                if nc_stop < stop_at:
                    stop_at = nc_stop

            while p < stop_at and (sum & mask):
                sum = _buzhash64_update(sum, p[0], p[window_size], window_size, self.table)
                p += 1

            did_bytes = p - (self.data + self.position)
            self.position += did_bytes
            self.remaining -= did_bytes

            if self.remaining <= window_size:
                self.fill()

        if self.remaining <= window_size:
            self.position += self.remaining
            self.remaining = 0

        old_last = self.last
        self.last = self.position
        n = self.last - old_last
        self.bytes_yielded += n

        # Return a memory view of the chunk
        return memoryview((self.data + old_last)[:n])

    def chunkify(self, fd, fh=-1, fmap=None):
        """
        Cut a file into chunks.

        :param fd: Python file object
        :param fh: OS-level file handle (if available),
                   defaults to -1 which means not to use OS-level fd.
        :param fmap: a file map, same format as generated by sparsemap
        """
        self._fd = fd
        self.fh = fh
        self.file_reader = FileReader(fd=fd, fh=fh, read_size=self.reader_block_size, sparse=self.sparse, fmap=fmap)
        self.done = 0
        self.remaining = 0
        self.bytes_read = 0
        self.bytes_yielded = 0
        self.position = 0
        self.last = 0
        self.eof = 0
        return self

    def __iter__(self):
        return self

    def __next__(self):
        started_chunking = time.monotonic()
        data = self.process()
        got = len(data)
        # we do not have SEEK_DATA/SEEK_HOLE support in chunker_process C code,
        # but we can just check if data was all-zero (and either came from a hole
        # or from stored zeros - we can not detect that here).
        if zeros.startswith(data):
            data = None
            allocation = CH_ALLOC
        else:
            allocation = CH_DATA
        self.chunking_time += time.monotonic() - started_chunking
        return Chunk(data, size=got, allocation=allocation)


def buzhash64(data, bytes key):
    cdef uint64_t *table
    cdef uint64_t sum
    table = buzhash64_init_table(key)
    sum = _buzhash64(<const unsigned char *> data, len(data), table)
    free(table)
    return sum


def buzhash64_update(uint64_t sum, unsigned char remove, unsigned char add, size_t len, bytes key):
    cdef uint64_t *table
    table = buzhash64_init_table(key)
    sum = _buzhash64_update(sum, remove, add, len, table)
    free(table)
    return sum


def buzhash64_get_table(bytes key):
    """Get the buzhash table generated from <key>."""
    cdef uint64_t *table
    cdef int i
    table = buzhash64_init_table(key)
    try:
        return [table[i] for i in range(256)]
    finally:
        free(table)



"""Tests for borg.legacy.helpers (borg1_hardlinkable, borg1_hardlink_master, borg1_hardlink_slave)."""

import stat

from ..item import Item
from ..legacy.helpers import borg1_hardlink_master, borg1_hardlink_slave, borg1_hardlinkable


def _item(**kwargs):
    base = {"path": "dir/file", "mode": stat.S_IFREG | 0o644, "mtime": 0}
    base.update(kwargs)
    return Item(internal_dict=base)


# ── borg1_hardlinkable ────────────────────────────────────────────────────────


def test_hardlinkable_regular_file():
    assert borg1_hardlinkable(stat.S_IFREG | 0o644)


def test_hardlinkable_block_device():
    assert borg1_hardlinkable(stat.S_IFBLK | 0o660)


def test_hardlinkable_char_device():
    assert borg1_hardlinkable(stat.S_IFCHR | 0o660)


def test_hardlinkable_fifo():
    assert borg1_hardlinkable(stat.S_IFIFO | 0o644)


def test_hardlinkable_symlink_is_false():
    assert not borg1_hardlinkable(stat.S_IFLNK | 0o777)


def test_hardlinkable_directory_is_false():
    assert not borg1_hardlinkable(stat.S_IFDIR | 0o755)


def test_hardlinkable_socket_is_false():
    assert not borg1_hardlinkable(stat.S_IFSOCK | 0o600)


# ── borg1_hardlink_master ─────────────────────────────────────────────────────


def test_hardlink_master_true_when_all_conditions_met():
    item = _item(hardlink_master=True)
    assert borg1_hardlink_master(item)


def test_hardlink_master_false_when_flag_missing():
    item = _item()
    assert not borg1_hardlink_master(item)


def test_hardlink_master_false_when_flag_is_false():
    item = _item(hardlink_master=False)
    assert not borg1_hardlink_master(item)


# ── borg1_hardlink_slave ──────────────────────────────────────────────────────


def test_hardlink_slave_true_when_source_and_hardlinkable():
    item = _item(source="dir/original")
    assert borg1_hardlink_slave(item)


def test_hardlink_slave_true_for_non_regular_hardlinkable_type():
    # borg1_hardlinkable covers FIFO/block/char as well as regular files
    item = _item(source="dir/original", mode=stat.S_IFIFO | 0o644)
    assert borg1_hardlink_slave(item)


def test_hardlink_slave_false_when_no_source():
    item = _item()
    assert not borg1_hardlink_slave(item)



"""Tests for borg.legacy.crypto.key (Pbkdf2FileMixin)."""

import pytest

from ...crypto.key import UnsupportedKeyFormatError
from ...helpers import msgpack
from ...legacy.crypto.key import AESCTRKey as LegacyAESCTRKey


# ── Pbkdf2FileMixin ───────────────────────────────────────────────────────────


def test_pbkdf2_encrypt_decrypt_roundtrip():
    # encrypt_key_file dispatches to encrypt_key_file_pbkdf2; decrypt_key_file
    # dispatches back — the round-trip must recover the original plaintext
    key_obj = LegacyAESCTRKey(None)
    plaintext = b"secret key material"
    blob = key_obj.encrypt_key_file(plaintext, "correct passphrase", "sha256")
    assert key_obj.decrypt_key_file(blob, "correct passphrase") == plaintext


def test_pbkdf2_wrong_passphrase_returns_none():
    # a wrong passphrase derives a different key, so the HMAC check fails;
    # decrypt_key_file signals this by returning None, not by raising
    key_obj = LegacyAESCTRKey(None)
    blob = key_obj.encrypt_key_file(b"secret key material", "correct passphrase", "sha256")
    assert key_obj.decrypt_key_file(blob, "wrong passphrase") is None


def test_pbkdf2_unsupported_version_raises():
    # only version 1 is defined in the borg 1.x format; anything else must raise
    blob = msgpack.packb({"version": 99})
    with pytest.raises(UnsupportedKeyFormatError):
        LegacyAESCTRKey(None).decrypt_key_file(blob, "pass")



"""Legacy RepoObj1 — Borg 1.x repository object format.

Moved from borg.repoobj as part of the legacy code separation.
"""

from ..constants import *  # NOQA
from ..helpers import workarounds
from ..compress import Compressor, get_compressor

# Workaround for lost passphrase or key in "authenticated" or "authenticated-blake2" mode
AUTHENTICATED_NO_KEY = "authenticated_no_key" in workarounds


class RepoObj1:  # legacy
    @classmethod
    def extract_crypted_data(cls, data: bytes) -> bytes:
        # used for crypto type detection
        return data

    def __init__(self, key):
        self.key = key
        self.compressor = get_compressor("lz4", legacy_mode=True)

    def id_hash(self, data: bytes) -> bytes:
        return self.key.id_hash(data)

    def format(
        self,
        id: bytes,
        meta: dict,
        data: bytes,
        compress: bool = True,
        size: int = None,
        ctype: int = None,
        clevel: int = None,
        ro_type: str = None,
    ) -> bytes:
        assert isinstance(id, bytes)
        assert meta == {}
        assert isinstance(data, (bytes, memoryview))
        assert ro_type is not None
        assert compress or size is not None and ctype is not None and clevel is not None
        if compress:
            assert size is None or size == len(data)
            meta, data_compressed = self.compressor.compress(meta, data)
        else:
            assert isinstance(size, int)
            data_compressed = data  # is already compressed, must include type/level bytes
        data_encrypted = self.key.encrypt(id, data_compressed)
        return data_encrypted

    def parse_meta(self, id: bytes, cdata: bytes) -> dict:
        raise NotImplementedError("parse_meta is not available for RepoObj1")

    def parse(
        self, id: bytes, cdata: bytes, decompress: bool = True, want_compressed: bool = False, ro_type: str = None
    ) -> tuple[dict, bytes]:
        assert not (not decompress and not want_compressed), "invalid parameter combination!"
        assert isinstance(id, bytes)
        assert isinstance(cdata, bytes)
        assert ro_type is not None
        data_compressed = self.key.decrypt(id, cdata)
        compressor_cls, compression_level = Compressor.detect(data_compressed[:2])
        compressor = compressor_cls(level=compression_level, legacy_mode=True)
        meta_compressed = {}
        meta_compressed["ctype"] = compressor.ID
        meta_compressed["clevel"] = compressor.level
        meta_compressed["csize"] = len(data_compressed)
        if decompress:
            meta, data = compressor.decompress(None, data_compressed)
            if not AUTHENTICATED_NO_KEY:
                self.key.assert_id(id, data)
        else:
            meta, data = None, None
        return meta_compressed if want_compressed else meta, data_compressed if want_compressed else data



import stat


def borg1_hardlinkable(mode):
    return stat.S_ISREG(mode) or stat.S_ISBLK(mode) or stat.S_ISCHR(mode) or stat.S_ISFIFO(mode)


def borg1_hardlink_master(item):
    return item.get("hardlink_master", False) and "source" not in item and borg1_hardlinkable(item.mode)


def borg1_hardlink_slave(item):
    return "source" in item and borg1_hardlinkable(item.mode)



# cython: language_level=3

import cython
import time

from cpython.bytes cimport PyBytes_AsString
from libc.stdint cimport uint8_t, uint64_t
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, memmove, memset

from ..crypto.low_level import CSPRNG

from ..constants import CH_DATA, CH_ALLOC, CH_HOLE, zeros
from .reader import FileReader, Chunk

# FastCDC content-defined chunker (Xia et al., USENIX ATC 2016).
#
# Differences vs. the buzhash64 chunker in this package:
#  * It uses the Gear rolling hash: fp = (fp << 1) + Gear[byte]. This is a single shift,
#    add and table lookup per byte (no window, no "remove" term), so it is cheaper than
#    buzhash's cyclic-polynomial update.
#  * The Gear table is keyed from a 256-bit key via the same CSPRNG used by buzhash64, so
#    cut points are unpredictable without the key (anti-fingerprinting), just like buzhash64.
#  * Because the Gear hash accumulates information in its HIGH bits (the low bits only depend
#    on the most recent bytes), the cut-decision mask uses the high bits of the hash.
#
# It implements the same FastCDC techniques the buzhash64 chunker uses: sub-minimum cut-point
# skipping, normalized chunking (strict/loose mask around a "normal" size), and min/max clamping.


@cython.boundscheck(False)
@cython.wraparound(False)
cdef uint64_t* fastcdc_init_gear(bytes key) except NULL:
    """Generate a keyed 256-entry, 64-bit Gear table deterministically from a 256-bit key."""
    rng = CSPRNG(key)
    cdef bytes rnd = rng.random_bytes(2048)  # 256 * sizeof(uint64_t)
    cdef const uint8_t* rp = <const uint8_t*>PyBytes_AsString(rnd)
    cdef uint64_t* gear = <uint64_t*>malloc(2048)
    if gear == NULL:
        raise MemoryError("Failed to allocate fastcdc gear table")
    cdef int i, j
    cdef uint64_t v
    for i in range(256):
        v = 0
        for j in range(8):
            v |= (<uint64_t>rp[i * 8 + j]) << (8 * j)
        gear[i] = v
    return gear


cdef inline uint64_t _high_mask(int bits):
    """A mask with <bits> one-bits in the most significant positions (Gear's strong bits)."""
    if bits <= 0:
        return 0
    if bits >= 64:
        return <uint64_t>0xFFFFFFFFFFFFFFFF
    return ((<uint64_t>1 << bits) - 1) << (64 - bits)


cdef class ChunkerFastCDC:
    """
    FastCDC content-defined chunker, variable chunk sizes, keyed Gear hash.

    Unlike the buzhash chunkers, Gear is window-less, so there is no hash_window_size parameter.
    """
    cdef uint64_t chunk_mask
    cdef uint64_t mask_s, mask_l  # normalized chunking: strict / loose masks
    cdef size_t normal_size       # chunk length at which we switch mask_s -> mask_l
    cdef int nc_level             # normalized chunking level (0 = disabled)
    cdef uint64_t* gear
    cdef uint8_t* data
    cdef object _fd
    cdef int fh
    cdef int done, eof
    cdef size_t min_size, buf_size, remaining, position, last
    cdef long long bytes_read, bytes_yielded
    cdef readonly float chunking_time
    cdef object file_reader
    cdef size_t reader_block_size
    cdef bint sparse

    def __cinit__(self, bytes key, int chunk_min_exp, int chunk_max_exp, int hash_mask_bits, int nc_level=0, size_t normal_size=0, bint sparse=False):
        self.gear = NULL
        self.data = NULL
        min_size = 1 << chunk_min_exp
        max_size = 1 << chunk_max_exp
        assert max_size <= len(zeros)
        assert min_size + 1 <= max_size, "too small max_size"

        self.chunk_mask = _high_mask(hash_mask_bits)
        self.min_size = min_size
        # Normalized chunking, identical structure to the buzhash64 chunker (see there), but with
        # the mask one-bits placed in the high bits of the Gear hash.
        assert nc_level >= 0
        assert hash_mask_bits - nc_level >= 1, "nc_level too large for hash_mask_bits"
        assert hash_mask_bits + nc_level <= 48, "nc_level too large for hash_mask_bits"
        self.nc_level = nc_level
        if nc_level:
            self.mask_s = _high_mask(hash_mask_bits + nc_level)
            self.mask_l = _high_mask(hash_mask_bits - nc_level)
            self.normal_size = normal_size if normal_size else ((1ULL << hash_mask_bits) - (1ULL << (hash_mask_bits - nc_level)))
        else:
            self.mask_s = self.chunk_mask
            self.mask_l = self.chunk_mask
            self.normal_size = 0
        self.gear = fastcdc_init_gear(key)
        self.buf_size = max_size
        self.data = <uint8_t*>malloc(self.buf_size)
        if self.data == NULL:
            raise MemoryError("Failed to allocate chunker buffer")
        self.fh = -1
        self.done = 0
        self.eof = 0
        self.remaining = 0
        self.position = 0
        self.last = 0
        self.bytes_read = 0
        self.bytes_yielded = 0
        self._fd = None
        self.chunking_time = 0.0
        self.reader_block_size = 1024 * 1024
        self.sparse = sparse

    def __dealloc__(self):
        if self.gear != NULL:
            free(self.gear)
            self.gear = NULL
        if self.data != NULL:
            free(self.data)
            self.data = NULL

    cdef int fill(self) except 0:
        """Fill the chunker's buffer with more data."""
        cdef ssize_t n
        cdef object chunk

        memmove(self.data, self.data + self.last, self.position + self.remaining - self.last)
        self.position -= self.last
        self.last = 0
        n = self.buf_size - self.position - self.remaining

        if self.eof or n == 0:
            return 1

        chunk = self.file_reader.read(n)
        n = chunk.meta["size"]

        if n > 0:
            if chunk.meta["allocation"] == CH_DATA:
                memcpy(self.data + self.position + self.remaining, <const unsigned char*>PyBytes_AsString(chunk.data), n)
            else:
                memset(self.data + self.position + self.remaining, 0, n)
            self.remaining += n
            self.bytes_read += n
        else:
            self.eof = 1
        return 1

    cdef object process(self) except *:
        """Process the chunker's buffer and return the next chunk."""
        cdef uint64_t fp = 0, mask, mask_s = self.mask_s, mask_l = self.mask_l
        cdef int nc_level = self.nc_level
        cdef size_t n, old_last, min_size = self.min_size
        cdef size_t normal_size = self.normal_size, normal_pos, chunk_len, did
        cdef uint8_t* p
        cdef uint8_t* stop
        cdef uint8_t* cut
        cdef uint64_t* gear = self.gear

        if self.done:
            if self.bytes_read == self.bytes_yielded:
                raise StopIteration
            else:
                raise Exception("chunkifier byte count mismatch")

        # ensure at least min_size + 1 bytes are buffered, or we are at eof
        while self.remaining < min_size + 1 and not self.eof:
            self.fill()

        # at eof with only a remainder (< min_size + 1): emit it as the final chunk
        if self.eof and self.remaining < min_size + 1:
            self.done = 1
            if self.remaining:
                old_last = self.last
                self.position += self.remaining
                self.last = self.position
                n = self.last - old_last
                self.remaining = 0
                self.bytes_yielded += n
                return memoryview((self.data + old_last)[:n])
            else:
                if self.bytes_read == self.bytes_yielded:
                    raise StopIteration
                else:
                    raise Exception("chunkifier byte count mismatch")

        # skip the sub-minimum region (no cut allowed below min_size), then gear-scan
        self.position += min_size
        self.remaining -= min_size
        fp = 0

        while True:
            chunk_len = self.position - self.last
            mask = mask_s if (nc_level and chunk_len < normal_size) else mask_l

            if self.remaining == 0:
                if self.eof:
                    break  # cut at end of data
                self.fill()
                if self.remaining == 0:
                    break  # buffer full -> chunk reached max_size -> forced cut
                continue

            p = self.data + self.position
            stop = p + self.remaining
            if nc_level and chunk_len < normal_size:
                # do not scan past the strict->loose transition; re-evaluate the mask there
                normal_pos = self.last + normal_size
                if (self.data + normal_pos) < stop:
                    stop = self.data + normal_pos

            cut = NULL
            while p < stop:
                fp = (fp << 1) + gear[p[0]]
                if (fp & mask) == 0:
                    cut = p
                    break
                p += 1

            if cut != NULL:
                p = cut + 1  # cut right after the byte that triggered the boundary
                did = p - (self.data + self.position)
                self.position += did
                self.remaining -= did
                break
            else:
                did = p - (self.data + self.position)
                self.position += did
                self.remaining -= did

        old_last = self.last
        self.last = self.position
        n = self.last - old_last
        self.bytes_yielded += n
        return memoryview((self.data + old_last)[:n])

    def chunkify(self, fd, fh=-1, fmap=None):
        self._fd = fd
        self.fh = fh
        self.file_reader = FileReader(fd=fd, fh=fh, read_size=self.reader_block_size, sparse=self.sparse, fmap=fmap)
        self.done = 0
        self.remaining = 0
        self.bytes_read = 0
        self.bytes_yielded = 0
        self.position = 0
        self.last = 0
        self.eof = 0
        return self

    def __iter__(self):
        return self

    def __next__(self):
        started_chunking = time.monotonic()
        data = self.process()
        got = len(data)
        if zeros.startswith(data):
            data = None
            allocation = CH_ALLOC
        else:
            allocation = CH_DATA
        self.chunking_time += time.monotonic() - started_chunking
        return Chunk(data, size=got, allocation=allocation)


def fastcdc_get_gear_table(bytes key):
    """Get the keyed gear table generated from <key> (for tests / inspection)."""
    cdef uint64_t* gear = fastcdc_init_gear(key)
    cdef int i
    try:
        return [gear[i] for i in range(256)]
    finally:
        free(gear)
